@@ -71,32 +71,32 @@ class KnotCollector(object):
         ctl.set_timeout(self._ttl)
         metric_families = dict()
 
-        def metric_families_append(family, labels, labels_val, data):
-            m = metric_families.get(family, GaugeMetricFamily(family, '', labels=labels))
-            c = metric_families.get(family + '_total', CounterMetricFamily(family, '', labels=labels))
+        def metric_families_append(family, labels, labels_val, data, unit=None):
+            m = metric_families.get(family, GaugeMetricFamily(family, '', labels=labels, unit=unit))
+            c = metric_families.get(family + '_total', CounterMetricFamily(family, '', labels=labels, unit=unit))
             m.add_metric(labels_val, data)
             c.add_metric(labels_val, data)
             metric_families[family] = m
             metric_families[family + '_total'] = c
 
         if self.collect_meminfo:
-            # Get global metrics.
             for pid, usage in memory_usage().items():
-                metric_families_append('knot_memory_usage', ['section', 'type'], ['server', str(pid)], usage)
+                metric_families_append('knot_memory_usage', ['pid'], [str(pid)], usage, "bytes")
 
         if self.collect_stats:
+            # Get global metrics.
             ctl.send_block(cmd="stats", flags="")
             global_stats = ctl.receive_stats()
 
             for section, section_data in global_stats.items():
                 for item, item_data in section_data.items():
-                    name = ('knot_' + item).replace('-', '_')
+                    name = ('knot_stats_' + item).replace('-', '_')
                     try:
                         for kind, kind_data in item_data.items():
-                            metric_families_append(name, ['section', 'type'], [section, kind], kind_data)
+                            metric_families_append(name, ['module', 'type'], [section, kind], kind_data)
 
                     except AttributeError:
-                        metric_families_append(name, ['section'], [section], item_data)
+                        metric_families_append(name, ['module', 'type'], [section, ""], item_data)
 
         if self.collect_zone_stats:
             # Get zone metrics.
@@ -107,12 +107,12 @@ class KnotCollector(object):
                 for zone, zone_data in zone_stats["zone"].items():
                     for section, section_data in zone_data.items():
                         for item, item_data in section_data.items():
-                            name = ('knot_' + item).replace('-', '_')
+                            name = ('knot_zone_stats_' + item).replace('-', '_')
                             try:
                                 for kind, kind_data in item_data.items():
-                                    metric_families_append(name, ['zone', 'section', 'type'], [zone, section, kind], kind_data)
+                                    metric_families_append(name, ['zone', 'module', 'type'], [zone, section, kind], kind_data)
                             except AttributeError:
-                                metric_families_append(name, ['zone', 'section'], [zone, section], item_data)
+                                metric_families_append(name, ['zone', 'module', 'type'], [zone, section, ""], item_data)
 
         if self.collect_zone_status:
             # zone state metrics
@@ -132,7 +132,7 @@ class KnotCollector(object):
                     if seconds == None:
                         continue
 
-                    metric_families_append('knot_zone_stats_' + metric, ['zone'], [zone], seconds)
+                    metric_families_append('knot_zone_status_' + metric, ['zone'], [zone], seconds, "seconds")
 
         if self.collect_zone_timers:
             # zone configuration metrics
@@ -149,7 +149,7 @@ class KnotCollector(object):
                 zone_config = params[name]['SOA']['data'][0].split(" ")
 
                 for metric in metrics:
-                    metric_families_append(metric['name'], ['zone'], [name], int(zone_config[metric['index']]))
+                    metric_families_append(metric['name'], ['zone'], [name], int(zone_config[metric['index']]), "seconds")
 
         for val in metric_families.values():
             yield val
@@ -158,9 +158,21 @@ class KnotCollector(object):
         ctl.close()
 
 
+class KnotHelpFormatter(argparse.HelpFormatter):
+    def _get_help_string(self, action):
+        help = action.help
+        if help is None:
+            help = ''
+
+        if not (type(action.default) is bool or action.default in (None, argparse.SUPPRESS)):
+            help += argparse._(' (default: %(default)s)')
+
+        return help
+
+
 def main():
     parser = argparse.ArgumentParser(
-        formatter_class = argparse.ArgumentDefaultsHelpFormatter,
+        formatter_class = KnotHelpFormatter,
     )
 
     parser.add_argument(
@@ -197,61 +209,58 @@ def main():
 
     parser.add_argument(
         "--no-meminfo",
-        action='store_false',
+        dest="meminfo",
+        action="store_false",
         help="disable collection of memory usage"
     )
 
     parser.add_argument(
         "--no-global-stats",
-        action='store_false',
+        dest="global_stats",
+        action="store_false",
         help="disable collection of global statistics"
     )
 
     parser.add_argument(
         "--no-zone-stats",
-        action='store_false',
+        dest="zone_stats",
+        action="store_false",
         help="disable collection of zone statistics"
     )
 
     parser.add_argument(
         "--no-zone-status",
-        action='store_false',
+        dest="zone_status",
+        action="store_false",
         help="disable collection of zone status"
     )
 
     parser.add_argument(
         "--no-zone-serial",
-        action='store_false',
+        dest="zone_serial",
+        action="store_false",
         help="disable collection of zone serial"
     )
 
     parser.add_argument(
         "--zone-timers",
-        action='store_true',
-        default=False,
+        dest="zone_timers",
+        action="store_true",
         help="enable collection of zone SOA timer values"
     )
 
-    parser.add_argument(
-        "--no-zone-timers",
-        action='store_const',
-        const=None,
-        help="supported for compatibility reasons; no effect",
-        # deprecated=True # in python >=3.13
-    )
-
     args = parser.parse_args()
 
     REGISTRY.register(KnotCollector(
         args.knot_library_path,
         args.knot_socket_path,
         args.knot_socket_timeout,
-        args.no_meminfo,
-        args.no_global_stats,
-        args.no_zone_stats,
-        args.no_zone_status,
+        args.meminfo,
+        args.global_stats,
+        args.zone_stats,
+        args.zone_status,
         args.zone_timers,
-        args.no_zone_serial,
+        args.zone_serial,
     ))
 
     class Server(http.server.HTTPServer):
