@@ -1,17 +1,6 @@
-/*  Copyright (C) 2024 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+/*  Copyright (C) CZ.NIC, z.s.p.o. and contributors
+ *  SPDX-License-Identifier: GPL-2.0-or-later
+ *  For more information, see <https://www.knot-dns.cz/>
  */
 
 #include <assert.h>
@@ -28,6 +17,7 @@
 
 conn_pool_t *global_conn_pool = NULL;
 conn_pool_t *global_sessticket_pool = NULL;
+conn_pool_t *global_redis_pool = NULL;
 
 const conn_pool_fd_t CONN_POOL_FD_INVALID = -1;
 
@@ -123,18 +113,23 @@ conn_pool_t *conn_pool_init(size_t capacity, knot_timediff_t timeout,
 	return pool;
 }
 
-void conn_pool_deinit(conn_pool_t *pool)
+void conn_pool_purge(conn_pool_t *pool)
 {
 	if (pool != NULL) {
-		pthread_cancel(pool->closing_thread);
-		pthread_join(pool->closing_thread, NULL);
-
 		conn_pool_fd_t fd;
 		knot_time_t unused;
 		while ((fd = get_old(pool, 0, &unused)) != CONN_POOL_FD_INVALID) {
 			pool->close_cb(fd);
 		}
+	}
+}
 
+void conn_pool_deinit(conn_pool_t *pool)
+{
+	if (pool != NULL) {
+		pthread_cancel(pool->closing_thread);
+		pthread_join(pool->closing_thread, NULL);
+		conn_pool_purge(pool);
 		pthread_mutex_destroy(&pool->mutex);
 		free(pool);
 	}
