@@ -110,12 +110,6 @@ ngtcp2_ssize ngtcp2_ppe_final(ngtcp2_ppe *ppe, const uint8_t **ppkt);
  */
 size_t ngtcp2_ppe_left(const ngtcp2_ppe *ppe);
 
-/*
- * ngtcp2_ppe_pktlen returns the provisional packet length.  It
- * includes AEAD overhead.
- */
-size_t ngtcp2_ppe_pktlen(const ngtcp2_ppe *ppe);
-
 /*
  * ngtcp2_ppe_dgram_padding is equivalent to call
  * ngtcp2_ppe_dgram_padding_size(ppe, NGTCP2_MAX_UDP_PAYLOAD_SIZE).
@@ -137,23 +131,17 @@ size_t ngtcp2_ppe_dgram_padding(ngtcp2_ppe *ppe);
  */
 size_t ngtcp2_ppe_dgram_padding_size(ngtcp2_ppe *ppe, size_t n);
 
-/*
- * ngtcp2_ppe_padding_hp_sample adds PADDING frame if the current
- * payload does not have enough space for header protection sample.
- * This function should be called just before calling
- * ngtcp2_ppe_final().
- *
- * This function returns the number of bytes added as padding.
- */
-size_t ngtcp2_ppe_padding_hp_sample(ngtcp2_ppe *ppe);
-
 /*
  * ngtcp2_ppe_padding_size adds PADDING frame so that the size of QUIC
- * packet is at least |n| bytes long.  If it is unable to add PADDING
- * in that way, this function still adds PADDING frame as much as
- * possible.  This function should be called just before calling
- * ngtcp2_ppe_final().  For Short packet, this function should be
- * called instead of ngtcp2_ppe_padding_hp_sample.
+ * packet is at least |n| bytes long and the current payload has
+ * enough space for header protection sample.  If it is unable to add
+ * PADDING at least |n| bytes, this function still adds PADDING frames
+ * as much as possible.  This function also adds PADDING frames so
+ * that the minimum padding requirement of header protection is met.
+ * Those padding may be larger than |n| bytes.  It is recommended to
+ * make sure that ngtcp2_ppe_ensure_hp_sample succeeds after writing
+ * QUIC packet header.  This function should be called just before
+ * calling ngtcp2_ppe_final().
  *
  * This function returns the number of bytes added as padding.
  */
