@@ -1,17 +1,6 @@
-/*  Copyright (C) 2024 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+/*  Copyright (C) CZ.NIC, z.s.p.o. and contributors
+ *  SPDX-License-Identifier: GPL-2.0-or-later
+ *  For more information, see <https://www.knot-dns.cz/>
  */
 
 #include <string.h>
@@ -661,8 +650,8 @@ static int finalize_previous_section(
 	};
 	knotd_conf_check_args_t args = {
 		.item = node->item,
-		.id = node->id,
-		.id_len = node->id_len,
+		.id = (node->parent != NULL) ? node->parent->id : node->id,
+		.id_len = (node->parent != NULL) ? node->parent->id_len : node->id_len,
 		.data = node->data,
 		.data_len = node->data_len,
 		.extra = &extra
@@ -697,8 +686,8 @@ static int finalize_item(
 	};
 	knotd_conf_check_args_t args = {
 		.item = (parser->event == YP_EID) ? node->item->var.g.id : node->item,
-		.id = node->id,
-		.id_len = node->id_len,
+		.id = (node->parent != NULL) ? node->parent->id : node->id,
+		.id_len = (node->parent != NULL) ? node->parent->id_len : node->id_len,
 		.data = node->data,
 		.data_len = node->data_len,
 		.extra = &extra
