@@ -1,17 +1,6 @@
-/*  Copyright (C) 2024 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+/*  Copyright (C) CZ.NIC, z.s.p.o. and contributors
+ *  SPDX-License-Identifier: GPL-2.0-or-later
+ *  For more information, see <https://www.knot-dns.cz/>
  */
 
 #include <assert.h>
@@ -442,6 +431,10 @@ size_t conf_val_count(
 		return 1;
 	}
 
+	if (val->blob_len == 0) { // Default value.
+		return 1;
+	}
+
 	size_t count = 0;
 	conf_val(val);
 	while (val->code == KNOT_EOK) {
@@ -464,26 +457,24 @@ void conf_val(
 	assert(val != NULL);
 	assert(val->code == KNOT_EOK || val->code == KNOT_EOF);
 
-	if (val->item->flags & YP_FMULTI) {
-		// Check if already called and not at the end.
-		if (val->data != NULL && val->code != KNOT_EOF) {
-			return;
-		}
-		// Otherwise set to the first value.
-		conf_val_reset(val);
-	} else {
-		// Check for empty data.
-		if (val->blob_len == 0) {
-			val->data = NULL;
-			val->len = 0;
-			val->code = KNOT_EOK;
-			return;
+	if (val->blob_len != 0) {
+		if (val->item->flags & YP_FMULTI) {
+			// Check if already called and not at the end.
+			if (val->data != NULL && val->code != KNOT_EOF) {
+				return;
+			}
+			// Otherwise set to the first value.
+			conf_val_reset(val);
 		} else {
-			assert(val->blob != NULL);
 			val->data = val->blob;
 			val->len = val->blob_len;
 			val->code = KNOT_EOK;
 		}
+	} else {
+		// Empty data.
+		val->data = NULL;
+		val->len = 0;
+		val->code = KNOT_EOK;
 	}
 }
 
@@ -1083,13 +1074,15 @@ static int str_label(
 	return str_zone(label, buff, buff_len);
 }
 
-static char* get_filename(
+char *conf_get_filename_txn(
 	conf_t *conf,
 	knot_db_txn_t *txn,
 	const knot_dname_t *zone,
 	const char *name)
 {
-	assert(name);
+	if (name == NULL) {
+		return NULL;
+	}
 
 	const char *end = name + strlen(name);
 	char out[1024] = "";
@@ -1200,7 +1193,7 @@ char* conf_zonefile_txn(
 		file = "%s.zone";
 	}
 
-	return get_filename(conf, txn, zone, file);
+	return conf_get_filename_txn(conf, txn, zone, file);
 }
 
 char* conf_db_txn(
@@ -1393,6 +1386,10 @@ conf_remote_t conf_remote_txn(
 	out.quic = conf_bool(&val);
 	val = conf_id_get_txn(conf, txn, C_RMT, C_TLS, id);
 	out.tls = conf_bool(&val);
+	val = conf_id_get_txn(conf, txn, C_RMT, C_NO_EDNS, id);
+	out.no_edns = conf_bool(&val);
+	val = conf_id_get_txn(conf, txn, C_RMT, C_BLOCK_NOTIFY_XFR, id);
+	out.block_notify_after_xfr = conf_bool(&val);
 
 	conf_val_t rundir_val = conf_get_txn(conf, txn, C_SRV, C_RUNDIR);
 	char *rundir = conf_abs_path(&rundir_val, NULL);
@@ -1428,8 +1425,7 @@ conf_remote_t conf_remote_txn(
 		conf_val_next(&val);
 	}
 
-	val = conf_id_get_txn(conf, txn, C_RMT, C_CERT_KEY, id);
-	out.pin = (uint8_t *)conf_bin(&val, &out.pin_len);
+	free(rundir);
 
 	// Get TSIG key (optional).
 	conf_val_t key_id = conf_id_get_txn(conf, txn, C_RMT, C_KEY, id);
@@ -1443,13 +1439,21 @@ conf_remote_t conf_remote_txn(
 		out.key.secret.data = (uint8_t *)conf_bin(&val, &out.key.secret.size);
 	}
 
-	free(rundir);
+	memset(out.hostnames, 0, sizeof(out.hostnames));
+	val = conf_id_get_txn(conf, txn, C_RMT, C_CERT_HOSTNAME, id);
+	for (size_t i = 0; val.code == KNOT_EOK; i++) {
+		out.hostnames[i] = conf_str(&val);
+		conf_val_next(&val);
+	}
 
-	val = conf_id_get_txn(conf, txn, C_RMT, C_BLOCK_NOTIFY_XFR, id);
-	out.block_notify_after_xfr = conf_bool(&val);
-
-	val = conf_id_get_txn(conf, txn, C_RMT, C_NO_EDNS, id);
-	out.no_edns = conf_bool(&val);
+	memset(out.pins, 0, sizeof(out.pins));
+	val = conf_id_get_txn(conf, txn, C_RMT, C_CERT_KEY, id);
+	for (size_t i = 0; val.code == KNOT_EOK; i++) {
+		size_t len;
+		out.pins[i] = (uint8_t *)conf_bin(&val, &len);
+		assert(len == KNOT_TLS_PIN_LEN);
+		conf_val_next(&val);
+	}
 
 	return out;
 }
@@ -1503,3 +1507,22 @@ int conf_xdp_iface(
 	return KNOT_EOK;
 #endif
 }
+
+bool conf_zone_rdb_enabled(
+	conf_t *conf,
+	const knot_dname_t *zone,
+	bool input,
+	uint8_t *instance)
+{
+	if (instance == NULL) {
+		return false;
+	}
+
+	conf_val_t val = conf_zone_get(conf, input ? C_ZONE_DB_IN : C_ZONE_DB_OUT, zone);
+	int64_t num = conf_int(&val);
+	if (num > 0) {
+		*instance = num;
+		return true;
+	}
+	return false;
+}
