@@ -1,17 +1,6 @@
-/*  Copyright (C) 2024 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+/*  Copyright (C) CZ.NIC, z.s.p.o. and contributors
+ *  SPDX-License-Identifier: GPL-2.0-or-later
+ *  For more information, see <https://www.knot-dns.cz/>
  */
 
 #pragma once
@@ -20,6 +9,7 @@
 
 #include "knot/conf/base.h"
 #include "knot/conf/schema.h"
+#include "libknot/quic/tls_common.h"
 
 /*! Configuration schema additional flags. */
 #define CONF_IO_FACTIVE		YP_FUSR1  /*!< Active confio transaction indicator. */
@@ -36,6 +26,8 @@
 #define CONF_IO_FRLD_ALL	(CONF_IO_FRLD_SRV | CONF_IO_FRLD_LOG | \
 				 CONF_IO_FRLD_MOD | CONF_IO_FRLD_ZONES)
 
+#define RMT_MAX_PINS		KNOT_TLS_MAX_PINS
+
 /*! Configuration remote getter output. */
 typedef struct {
 	/*! Target socket address. */
@@ -46,16 +38,16 @@ typedef struct {
 	bool quic;
 	/*! TLS context. */
 	bool tls;
-	/*! TSIG key. */
-	knot_tsig_key_t key;
-	/*! Suppress sending NOTIFY after zone transfer from this master. */
-	bool block_notify_after_xfr;
 	/*! Disable EDNS on XFR queries. */
 	bool no_edns;
-	/*! Possible remote certificate PIN. */
-	const uint8_t *pin;
-	/*! Length of the remote certificate PIN. Zero if PIN not specified. */
-	size_t pin_len;
+	/*! Suppress sending NOTIFY after zone transfer from this master. */
+	bool block_notify_after_xfr;
+	/*! TSIG key. */
+	knot_tsig_key_t key;
+	/*! Remote certificate permittable hostnames. */
+	const char *hostnames[RMT_MAX_PINS];
+	/*! Remote certificate permittable PINs. */
+	const uint8_t *pins[RMT_MAX_PINS];
 } conf_remote_t;
 
 /*! Configuration section iterator. */
@@ -672,6 +664,30 @@ void conf_free_mod_id(
 	conf_mod_id_t *mod_id
 );
 
+/*!
+ * Get the absolute path in zonefile-like manner for given name/template.
+ *
+ * \param[in] conf    Configuration.
+ * \param[in] txn     Configuration DB transaction.
+ * \param[in] zone    Zone name.
+ * \param[in] name    File name or template (like %s.zonediff).
+ *
+ * \return Absolute zone file path string pointer.
+ */
+char *conf_get_filename_txn(
+	conf_t *conf,
+	knot_db_txn_t *txn,
+	const knot_dname_t *zone,
+	const char *name
+);
+static inline char* conf_get_filename(
+	conf_t *conf,
+	const knot_dname_t *zone,
+	const char *name)
+{
+	return conf_get_filename_txn(conf, &conf->read_txn, zone, name);
+}
+
 /*!
  * Gets the absolute zone file path.
  *
@@ -975,3 +991,21 @@ int conf_xdp_iface(
 	struct sockaddr_storage *addr,
 	conf_xdp_iface_t *iface
 );
+
+/*!
+ * Checks if zone database backend is enabled for the zone.
+ *
+ * \param[in] conf       Configuration.
+ * \param[in] zone       Zone name.
+ * \param[in] input      Input backend indication; Output backend otherwise.
+ * \param[out] instance  Configured instance number.
+ *
+ * \retval true   if enabled.
+ * \retval false  if disabled.
+ */
+bool conf_zone_rdb_enabled(
+	conf_t *conf,
+	const knot_dname_t *zone,
+	bool input,
+	uint8_t *instance
+);
