@@ -1,17 +1,6 @@
-/*  Copyright (C) 2024 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+/*  Copyright (C) CZ.NIC, z.s.p.o. and contributors
+ *  SPDX-License-Identifier: GPL-2.0-or-later
+ *  For more information, see <https://www.knot-dns.cz/>
  */
 
 #pragma once
@@ -41,9 +30,11 @@
 #define C_CATALOG_ROLE		"\x0C""catalog-role"
 #define C_CATALOG_TPL		"\x10""catalog-template"
 #define C_CATALOG_ZONE		"\x0C""catalog-zone"
+#define C_CA_FILE		"\x07""ca-file"
 #define C_CDS_CDNSKEY		"\x13""cds-cdnskey-publish"
 #define C_CDS_DIGESTTYPE	"\x0F""cds-digest-type"
 #define C_CERT_FILE		"\x09""cert-file"
+#define C_CERT_HOSTNAME		"\x0D""cert-hostname"
 #define C_CERT_KEY		"\x08""cert-key"
 #define C_CHK_INTERVAL		"\x0E""check-interval"
 #define C_CLEAR			"\x05""clear"
@@ -64,13 +55,19 @@
 #define C_DNSSEC_VALIDATION	"\x11""dnssec-validation"
 #define C_DOMAIN		"\x06""domain"
 #define C_DS_PUSH		"\x07""ds-push"
+#define C_DUMP_ADD		"\x0E""dump-additions"
+#define C_DUMP_NEW		"\x0D""dump-new-zone"
+#define C_DUMP_REM		"\x0D""dump-removals"
 #define C_ECS			"\x12""edns-client-subnet"
 #define C_EXPIRE_MAX_INTERVAL	"\x13""expire-max-interval"
 #define C_EXPIRE_MIN_INTERVAL	"\x13""expire-min-interval"
+#define C_EXTERNAL		"\x08""external"
+#define C_EXTERNAL_VLDT		"\x13""external-validation"
 #define C_FILE			"\x04""file"
 #define C_GLOBAL_MODULE		"\x0D""global-module"
 #define C_ID			"\x02""id"
 #define C_IDENT			"\x08""identity"
+#define C_INCLUDE_FROM		"\x0C""include-from"
 #define C_INCL			"\x07""include"
 #define C_IXFR_BENEVOLENT	"\x0F""ixfr-benevolent"
 #define C_IXFR_BY_ONE		"\x0B""ixfr-by-one"
@@ -90,6 +87,7 @@
 #define C_KEY_FILE		"\x08""key-file"
 #define C_KEY_LABEL		"\x09""key-label"
 #define C_KSK_LIFETIME		"\x0C""ksk-lifetime"
+#define C_KSK_ONLY		"\x08""ksk-only"
 #define C_KSK_SBM		"\x0E""ksk-submission"
 #define C_KSK_SHARED		"\x0a""ksk-shared"
 #define C_KSK_SIZE		"\x08""ksk-size"
@@ -179,6 +177,7 @@
 #define C_UDP_MAX_PAYLOAD_IPV6	"\x14""udp-max-payload-ipv6"
 #define C_UDP_WORKERS		"\x0B""udp-workers"
 #define C_UNSAFE_OPERATION	"\x10""unsafe-operation"
+#define C_UPDATE_DELAY		"\x0C""update-delay"
 #define C_UPDATE_OWNER		"\x0C""update-owner"
 #define C_UPDATE_OWNER_MATCH	"\x12""update-owner-match"
 #define C_UPDATE_OWNER_NAME	"\x11""update-owner-name"
@@ -189,9 +188,16 @@
 #define C_XDP			"\x03""xdp"
 #define C_ZONE			"\x04""zone"
 #define C_ZONEFILE_LOAD		"\x0D""zonefile-load"
+#define C_ZONEFILE_SKIP		"\x0D""zonefile-skip"
 #define C_ZONEFILE_SYNC		"\x0D""zonefile-sync"
 #define C_ZONEMD_GENERATE	"\x0F""zonemd-generate"
 #define C_ZONEMD_VERIFY		"\x0D""zonemd-verify"
+#define C_ZONE_DB_CERT_HOSTNAME	"\x15""zone-db-cert-hostname"
+#define C_ZONE_DB_CERT_KEY	"\x10""zone-db-cert-key"
+#define C_ZONE_DB_IN		"\x0D""zone-db-input"
+#define C_ZONE_DB_LISTEN	"\x0E""zone-db-listen"
+#define C_ZONE_DB_OUT		"\x0E""zone-db-output"
+#define C_ZONE_DB_TLS		"\x0B""zone-db-tls"
 #define C_ZONE_MAX_SIZE		"\x0D""zone-max-size"
 #define C_ZONE_MAX_TTL		"\x0C""zone-max-ttl"
 #define C_ZSK_LIFETIME		"\x0C""zsk-lifetime"
@@ -272,9 +278,10 @@ enum {
 	DBUS_EVENT_NONE            = 0,
 	DBUS_EVENT_RUNNING         = (1 << 0),
 	DBUS_EVENT_ZONE_UPDATED    = (1 << 1),
-	DBUS_EVENT_ZONE_SUBMISSION = (1 << 2),
-	DBUS_EVENT_ZONE_INVALID    = (1 << 3),
-	DBUS_EVENT_KEYS_UPDATED    = (1 << 4),
+	DBUS_EVENT_EXTERNAL        = (1 << 2),
+	DBUS_EVENT_ZONE_SUBMISSION = (1 << 3),
+	DBUS_EVENT_ZONE_INVALID    = (1 << 4),
+	DBUS_EVENT_KEYS_UPDATED    = (1 << 5),
 };
 
 extern const knot_lookup_t acl_actions[];
