@@ -1,17 +1,6 @@
-/*  Copyright (C) 2025 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+/*  Copyright (C) CZ.NIC, z.s.p.o. and contributors
+ *  SPDX-License-Identifier: GPL-2.0-or-later
+ *  For more information, see <https://www.knot-dns.cz/>
  */
 
 #include <fnmatch.h>
@@ -40,8 +29,10 @@
 #include "knot/conf/module.h"
 #include "knot/conf/schema.h"
 #include "knot/common/log.h"
+#include "knot/ctl/process.h"
 #include "knot/updates/acl.h"
 #include "knot/zone/serial.h"
+#include "knot/zone/skip.h"
 #include "libknot/errcode.h"
 #include "libknot/quic/tls_common.h"
 #include "libknot/yparser/yptrafo.h"
@@ -382,6 +373,31 @@ int check_modulo_shift(
 	return KNOT_EOK;
 }
 
+int check_zonefile_skip(
+	knotd_conf_check_args_t *args)
+{
+	zone_skip_t skip = { 0 };
+	if (zone_skip_add(&skip, (const char *)args->data) != KNOT_EOK) {
+		args->err_str = "invalid type";
+		return KNOT_EINVAL;
+	}
+	zone_skip_free(&skip);
+	return KNOT_EOK;
+}
+
+int check_ctl_listen(
+	knotd_conf_check_args_t *args)
+{
+	conf_val_t val = conf_get_txn(args->extra->conf, args->extra->txn,
+	                              C_CTL, C_LISTEN);
+	if (conf_val_count(&val) > CTL_MAX_CONCURRENT / 2) {
+		args->err_str = "too many control sockets configured";
+		return KNOT_EINVAL;
+	}
+
+	return KNOT_EOK;
+}
+
 static int dir_exists(const char *dir)
 {
 	struct stat st;
@@ -743,8 +759,6 @@ int check_policy(
 	                                      C_NSEC3, args->id, args->id_len);
 	conf_val_t nsec3_iters = conf_rawid_get_txn(args->extra->conf, args->extra->txn, C_POLICY,
 	                                            C_NSEC3_ITER, args->id, args->id_len);
-	conf_val_t nsec3_salt_len = conf_rawid_get_txn(args->extra->conf, args->extra->txn, C_POLICY,
-	                                               C_NSEC3_SALT_LEN, args->id, args->id_len);
 
 	unsigned algorithm = conf_opt(&alg);
 
@@ -815,11 +829,6 @@ int check_policy(
 			CONF_LOG(LOG_NOTICE, "policy[%s].nsec3-iterations=%u is too high, "
 			                     "the recommended value is 0", args->id, iters);
 		}
-		uint16_t salt_len = conf_int(&nsec3_salt_len);
-		if (nsec3_salt_len.code != KNOT_EOK) {
-			CONF_LOG(LOG_NOTICE, "policy '%s' depends on default nsec3-salt-length=%i, "
-			                     "since version 3.5 the default becomes 0", args->id, salt_len);
-		}
 	}
 
 	conf_val_t dnskey_mgmt = conf_rawid_get_txn(args->extra->conf, args->extra->txn, C_POLICY,
@@ -842,6 +851,12 @@ int check_policy(
 	return KNOT_EOK;
 }
 
+int check_external(
+	knotd_conf_check_args_t *args)
+{
+	return KNOT_EOK;
+}
+
 int check_key(
 	knotd_conf_check_args_t *args)
 {
@@ -864,11 +879,14 @@ int check_acl(
 	                                    C_KEY, args->id, args->id_len);
 	conf_val_t proto = conf_rawid_get_txn(args->extra->conf, args->extra->txn, C_ACL,
 	                                      C_PROTOCOL, args->id, args->id_len);
+	conf_val_t hostname = conf_rawid_get_txn(args->extra->conf, args->extra->txn, C_ACL,
+	                                         C_CERT_HOSTNAME, args->id, args->id_len);
 	conf_val_t remote = conf_rawid_get_txn(args->extra->conf, args->extra->txn, C_ACL,
 	                                       C_RMT, args->id, args->id_len);
 	if (remote.code != KNOT_ENOENT &&
-	    (addr.code != KNOT_ENOENT || key.code != KNOT_ENOENT || proto.code != KNOT_ENOENT)) {
-		args->err_str = "specified ACL/remote together with address, key, or protocol";
+	    (addr.code != KNOT_ENOENT || key.code != KNOT_ENOENT ||
+	     proto.code != KNOT_ENOENT || hostname.code != KNOT_ENOENT)) {
+		args->err_str = "specified ACL/remote together with address, key, protocol, or hostname";
 		return KNOT_EINVAL;
 	}
 
@@ -919,6 +937,19 @@ int check_remote(
 #endif
 	}
 
+	conf_val_t pin = conf_rawid_get_txn(args->extra->conf, args->extra->txn, C_RMT,
+	                                    C_CERT_KEY, args->id, args->id_len);
+	if (conf_val_count(&pin) > RMT_MAX_PINS) {
+		args->err_str = "too many cert-keys";
+		return KNOT_EINVAL;
+	}
+	conf_val_t cert_host = conf_rawid_get_txn(args->extra->conf, args->extra->txn, C_RMT,
+	                                          C_CERT_HOSTNAME, args->id, args->id_len);
+	if (conf_val_count(&cert_host) > RMT_MAX_PINS) {
+		args->err_str = "too many cert-hosts";
+		return KNOT_EINVAL;
+	}
+
 	return KNOT_EOK;
 }
 
@@ -994,6 +1025,18 @@ static conf_val_t conf_get_wrap(
 	}
 }
 
+int check_include_from(
+	knotd_conf_check_args_t *args)
+{
+	if (knot_dname_in_bailiwick(args->data, args->id) < 0 ||
+	    knot_dname_is_equal(args->data, args->id)) {
+		args->err_str = "not a subzone";
+		return KNOT_EINVAL;
+	}
+
+	return KNOT_EOK;
+}
+
 #define CHECK_ZONE_INTERVALS(low_item, high_item) { \
 	conf_val_t high = conf_get_wrap(args, high_item); \
 	if (high.code == KNOT_EOK) { \
@@ -1146,6 +1189,15 @@ static int check_zone_or_tpl(
 		}
 	}
 
+	conf_val_t inc_from = conf_get_wrap(args, C_INCLUDE_FROM);
+	if (inc_from.code == KNOT_EOK) {
+		conf_val_t rev_from = conf_get_wrap(args, C_REVERSE_GEN);
+		if (rev_from.code == KNOT_EOK) {
+			args->err_str = "include-from not compatible with reverse-from";
+			return KNOT_EINVAL;
+		}
+	}
+
 	return KNOT_EOK;
 }
 
@@ -1161,6 +1213,29 @@ int check_catalog_tpl(
 	return check_zone_or_tpl(args);
 }
 
+int check_rdb(
+	knotd_conf_check_args_t *args)
+{
+#ifndef ENABLE_REDIS
+	args->err_str = "Zone database support not available";
+	return KNOT_ENOTSUP;
+#else
+	return KNOT_EOK;
+#endif
+}
+
+int check_db_instance(
+	knotd_conf_check_args_t *args)
+{
+	int64_t instance = yp_int(args->data);
+	if (instance == 0) {
+		args->err_str = "instance number must be 1-8";
+		return KNOT_EINVAL;
+	}
+
+	return KNOT_EOK;
+}
+
 static int glob_error(
 	const char *epath,
 	int eerrno)
