@@ -1,17 +1,6 @@
-/*  Copyright (C) 2025 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+/*  Copyright (C) CZ.NIC, z.s.p.o. and contributors
+ *  SPDX-License-Identifier: GPL-2.0-or-later
+ *  For more information, see <https://www.knot-dns.cz/>
  */
 
 #pragma once
@@ -94,6 +83,14 @@ int check_modulo_shift(
 	knotd_conf_check_args_t *args
 );
 
+int check_zonefile_skip(
+	knotd_conf_check_args_t *args
+);
+
+int check_ctl_listen(
+	knotd_conf_check_args_t *args
+);
+
 int check_database(
 	knotd_conf_check_args_t *args
 );
@@ -122,6 +119,10 @@ int check_policy(
 	knotd_conf_check_args_t *args
 );
 
+int check_external(
+	knotd_conf_check_args_t *args
+);
+
 int check_key(
 	knotd_conf_check_args_t *args
 );
@@ -150,6 +151,10 @@ int check_template(
 	knotd_conf_check_args_t *args
 );
 
+int check_include_from(
+	knotd_conf_check_args_t *args
+);
+
 int check_zone(
 	knotd_conf_check_args_t *args
 );
@@ -158,6 +163,14 @@ int check_catalog_tpl(
 	knotd_conf_check_args_t *args
 );
 
+int check_rdb(
+	knotd_conf_check_args_t *args
+);
+
+int check_db_instance(
+	knotd_conf_check_args_t *args
+);
+
 int include_file(
 	knotd_conf_check_args_t *args
 );
