@@ -1,17 +1,6 @@
-/*  Copyright (C) 2025 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+/*  Copyright (C) CZ.NIC, z.s.p.o. and contributors
+ *  SPDX-License-Identifier: GPL-2.0-or-later
+ *  For more information, see <https://www.knot-dns.cz/>
  */
 
 #include <string.h>
@@ -72,7 +61,7 @@ static struct {
 	            sizeof(((send_ctx_t *)0)->ttl) +
 	            sizeof(((send_ctx_t *)0)->type) +
 	            sizeof(((send_ctx_t *)0)->rdata)];
-} ctl_globals[CTL_MAX_CONCURRENT + 1];
+} ctl_globals[CTL_MAX_CONCURRENT];
 
 static bool allow_blocking_while_ctl_txn(zone_event_type_t event)
 {
@@ -325,7 +314,12 @@ static int zone_status(zone_t *zone, ctl_args_t *args)
 
 	if (MATCH_OR_FILTER(args, CTL_FILTER_STATUS_TRANSACTION)) {
 		data[KNOT_CTL_IDX_TYPE] = "transaction";
-		data[KNOT_CTL_IDX_DATA] = (zone->control_update != NULL) ? "open" : STATUS_EMPTY;
+		const char *value = STATUS_EMPTY;
+		if (zone->control_update != NULL) {
+			value = (zone->control_update->flags & UPDATE_WFEV) ?
+			        "open-external" : "open";
+		}
+		data[KNOT_CTL_IDX_DATA] = value;
 		ret = knot_ctl_send(args->ctl, type, &data);
 		if (ret != KNOT_EOK) {
 			return ret;
@@ -700,7 +694,7 @@ static int zone_backup_cmd(zone_t *zone, ctl_args_t *args)
 
 	int ret = KNOT_EOK;
 	pthread_mutex_lock(&zone->cu_lock);
-	if (zone->backup_ctx != NULL) {
+	if (ATOMIC_GET(zone->backup_ctx) != NULL) {
 		log_zone_warning(zone->name, "backup or restore already in progress, skipping zone");
 		ctx->failed = true;
 		ret = KNOT_EPROGRESS;
@@ -713,7 +707,7 @@ static int zone_backup_cmd(zone_t *zone, ctl_args_t *args)
 	}
 
 	if (ret == KNOT_EOK) {
-		zone->backup_ctx = ctx;
+		ATOMIC_SET(zone->backup_ctx, ctx);
 	}
 	pthread_mutex_unlock(&zone->cu_lock);
 
@@ -737,7 +731,7 @@ static int zone_backup_cmd(zone_t *zone, ctl_args_t *args)
 	}
 
 	if (ret != KNOT_EOK || finish) {
-		zone->backup_ctx = NULL;
+		ATOMIC_SET(zone->backup_ctx, NULL);
 		return ret;
 	}
 
@@ -836,9 +830,9 @@ static int zone_key_roll(zone_t *zone, ctl_args_t *args)
 	}
 
 	const char *key_type = args->data[KNOT_CTL_IDX_TYPE];
-	if (strncasecmp(key_type, "ksk", 3) == 0) {
+	if (strncasecmp(key_type, CMD_ROLLOVER_KSK, 3) == 0) {
 		zone_set_flag(zone, ZONE_FORCE_KSK_ROLL);
-	} else if (strncasecmp(key_type, "zsk", 3) == 0) {
+	} else if (strncasecmp(key_type, CMD_ROLLOVER_ZSK, 3) == 0) {
 		zone_set_flag(zone, ZONE_FORCE_ZSK_ROLL);
 	} else {
 		return KNOT_EINVAL;
@@ -902,7 +896,7 @@ static int zone_txn_begin_l(zone_t *zone, _unused_ ctl_args_t *args)
 		return KNOT_TXN_EEXISTS;
 	}
 
-	struct zone_backup_ctx *backup_ctx = zone->backup_ctx;
+	struct zone_backup_ctx *backup_ctx = ATOMIC_GET(zone->backup_ctx);
 	if (backup_ctx != NULL && backup_ctx->restore_mode) {
 		log_zone_warning(zone->name, "zone restore pending, try opening control transaction later");
 		return KNOT_ETRYAGAIN;
@@ -944,16 +938,19 @@ static int zone_txn_commit_l(zone_t *zone, _unused_ ctl_args_t *args)
 		return KNOT_TXN_ENOTEXISTS;
 	}
 
+	if (zone->control_update->flags & UPDATE_WFEV) {
+		zone->control_update->flags |= UPDATE_EVOK;
+		knot_sem_post(&zone->control_update->external);
+		return KNOT_EOK;
+	}
+
 	int ret = zone_update_semcheck(conf(), zone->control_update);
 	if (ret != KNOT_EOK) {
 		return ret; // Recoverable error.
 	}
 
 	// NOOP if empty changeset/contents.
-	if (((zone->control_update->flags & UPDATE_INCREMENTAL) &&
-	     changeset_empty(&zone->control_update->change)) ||
-	    ((zone->control_update->flags & UPDATE_FULL) &&
-	     zone_contents_is_empty(zone->control_update->new_cont))) {
+	if (zone_update_no_change(zone->control_update)) {
 		zone_control_clear(zone);
 		return KNOT_EOK;
 	}
@@ -1016,6 +1013,12 @@ static int zone_txn_abort(zone_t *zone, _unused_ ctl_args_t *args)
 		return KNOT_TXN_ENOTEXISTS;
 	}
 
+	if (zone->control_update->flags & UPDATE_WFEV) {
+		knot_sem_post(&zone->control_update->external);
+		pthread_mutex_unlock(&zone->cu_lock);
+		return KNOT_EOK;
+	}
+
 	zone_control_clear(zone);
 
 	pthread_mutex_unlock(&zone->cu_lock);
@@ -1060,7 +1063,7 @@ static int init_send_ctx(send_ctx_t *ctx, const knot_dname_t *zone_name,
 	return KNOT_EOK;
 }
 
-static int send_rrset(knot_rrset_t *rrset, send_ctx_t *ctx)
+static int send_rrset(const knot_rrset_t *rrset, send_ctx_t *ctx)
 {
 	if (rrset->type != KNOT_RRTYPE_RRSIG) {
 		int ret = snprintf(ctx->ttl, sizeof(ctx->ttl), "%u", rrset->ttl);
@@ -1097,6 +1100,16 @@ static int send_rrset(knot_rrset_t *rrset, send_ctx_t *ctx)
 	return KNOT_EOK;
 }
 
+static int send_rrset_callback(const knot_rrset_t *rrset, void *ctx_void)
+{
+	send_ctx_t *ctx = ctx_void;
+	char *owner = knot_dname_to_str(ctx->owner, rrset->owner, sizeof(ctx->owner));
+	if (owner == NULL) {
+		return KNOT_EINVAL;
+	}
+	return send_rrset(rrset, ctx);
+}
+
 static int send_node(zone_node_t *node, void *ctx_void)
 {
 	send_ctx_t *ctx = ctx_void;
@@ -1252,66 +1265,6 @@ static int zone_txn_get(zone_t *zone, ctl_args_t *args)
 	return ret;
 }
 
-static int send_changeset_part(changeset_t *ch, send_ctx_t *ctx, bool from)
-{
-	ctx->data[KNOT_CTL_IDX_FILTERS] = from ? CTL_FILTER_DIFF_REM_R : CTL_FILTER_DIFF_ADD_R;
-
-	// Send SOA only if explicitly changed.
-	if (ch->soa_to != NULL) {
-		knot_rrset_t *soa = from ? ch->soa_from : ch->soa_to;
-		assert(soa);
-
-		char *owner = knot_dname_to_str(ctx->owner, soa->owner, sizeof(ctx->owner));
-		if (owner == NULL) {
-			return KNOT_EINVAL;
-		}
-
-		int ret = send_rrset(soa, ctx);
-		if (ret != KNOT_EOK) {
-			return ret;
-		}
-	}
-
-	// Send other records.
-	changeset_iter_t it;
-	int ret = from ? changeset_iter_rem(&it, ch) : changeset_iter_add(&it, ch);
-	if (ret != KNOT_EOK) {
-		return ret;
-	}
-
-	knot_rrset_t rrset = changeset_iter_next(&it);
-	while (!knot_rrset_empty(&rrset)) {
-		char *owner = knot_dname_to_str(ctx->owner, rrset.owner, sizeof(ctx->owner));
-		if (owner == NULL) {
-			changeset_iter_clear(&it);
-			return KNOT_EINVAL;
-		}
-
-		ret = send_rrset(&rrset, ctx);
-		if (ret != KNOT_EOK) {
-			changeset_iter_clear(&it);
-			return ret;
-		}
-
-		rrset = changeset_iter_next(&it);
-	}
-	changeset_iter_clear(&it);
-
-	return KNOT_EOK;
-}
-
-static int send_changeset(changeset_t *ch, send_ctx_t *ctx)
-{
-	// First send 'from' changeset part.
-	int ret = send_changeset_part(ch, ctx, true);
-	if (ret != KNOT_EOK) {
-		return ret;
-	}
-
-	// Second send 'to' changeset part.
-	return send_changeset_part(ch, ctx, false);
-}
-
 static int zone_txn_diff_l(zone_t *zone, ctl_args_t *args)
 {
 	if (zone->control_update == NULL) {
@@ -1326,11 +1279,16 @@ static int zone_txn_diff_l(zone_t *zone, ctl_args_t *args)
 
 	send_ctx_t *ctx = &ctl_globals[args->thread_idx].send_ctx;
 	int ret = init_send_ctx(ctx, zone->name, args);
-	if (ret != KNOT_EOK) {
-		return ret;
+	if (ret == KNOT_EOK) {
+		ctx->data[KNOT_CTL_IDX_FILTERS] = CTL_FILTER_DIFF_REM_R;
+		ret = zone_update_foreach(zone->control_update, false, send_rrset_callback, ctx);
+	}
+	if (ret == KNOT_EOK) {
+		ctx->data[KNOT_CTL_IDX_FILTERS] = CTL_FILTER_DIFF_ADD_R;
+		ret = zone_update_foreach(zone->control_update, true, send_rrset_callback, ctx);
 	}
 
-	return send_changeset(&zone->control_update->change, ctx);
+	return ret;
 }
 
 static int zone_txn_diff(zone_t *zone, ctl_args_t *args)
@@ -1964,7 +1922,7 @@ static int ctl_zone(ctl_args_t *args, ctl_cmd_t cmd)
 
 static void check_zone_txn(zone_t *zone, const knot_dname_t **exists)
 {
-	if (zone->control_update != NULL) {
+	if (zone->control_update != NULL && !(zone->control_update->flags & UPDATE_WFEV)) {
 		*exists = zone->name;
 	}
 }
@@ -1985,43 +1943,44 @@ static int check_no_zone_txn(server_t *server, const char *action)
 
 static int server_status(ctl_args_t *args)
 {
-	const char *type = args->data[KNOT_CTL_IDX_TYPE];
-
-	if (type == NULL || strlen(type) == 0) {
-		return KNOT_EOK;
-	}
-
 	char buff[4096] = "";
 
-	int ret;
-	if (strcasecmp(type, "version") == 0) {
-		ret = snprintf(buff, sizeof(buff), "%s", PACKAGE_VERSION);
-	} else if (strcasecmp(type, "workers") == 0) {
-		int running_bkg_wrk, wrk_queue;
-		worker_pool_status(args->server->workers, false, &running_bkg_wrk, &wrk_queue);
-		ret = snprintf(buff, sizeof(buff), "UDP workers: %zu, TCP workers: %zu, "
-		               "XDP workers: %zu, background workers: %zu (running: %d, pending: %d)",
-		               conf()->cache.srv_udp_threads, conf()->cache.srv_tcp_threads,
-		               conf()->cache.srv_xdp_threads, conf()->cache.srv_bg_threads,
-		               running_bkg_wrk, wrk_queue);
-	} else if (strcasecmp(type, "configure") == 0) {
-		ret = snprintf(buff, sizeof(buff), "%s", configure_summary);
-	} else if (strcasecmp(type, "cert-key") == 0) {
-		uint8_t pin[128];
-		size_t pin_len = server_cert_pin(args->server, pin, sizeof(pin));
-		if (pin_len > 0) {
-			ret = snprintf(buff, sizeof(buff), "%.*s", (int)pin_len, pin);
-		} else {
-			ret = snprintf(buff, sizeof(buff), STATUS_EMPTY);
+	const char *type = args->data[KNOT_CTL_IDX_TYPE];
+	if (type == NULL || strlen(type) == 0) {
+		if (!(args->server->state & ServerAnswering)) {
+			args->data[KNOT_CTL_IDX_FILTERS] = CTL_FILTER_STATUS_LOADING;
 		}
 	} else {
-		return KNOT_EINVAL;
-	}
-	if (ret <= 0 || ret >= sizeof(buff)) {
-		return KNOT_ESPACE;
-	}
+		int ret;
+		if (strcasecmp(type, CMD_STATUS_VERSION) == 0) {
+			ret = snprintf(buff, sizeof(buff), "%s", PACKAGE_VERSION);
+		} else if (strcasecmp(type, CMD_STATUS_WORKERS) == 0) {
+			int running_bkg_wrk, wrk_queue;
+			worker_pool_status(args->server->workers, false, &running_bkg_wrk, &wrk_queue);
+			ret = snprintf(buff, sizeof(buff), "UDP workers: %zu, TCP workers: %zu, "
+			               "XDP workers: %zu, background workers: %zu (running: %d, pending: %d)",
+			               conf()->cache.srv_udp_threads, conf()->cache.srv_tcp_threads,
+			               conf()->cache.srv_xdp_threads, conf()->cache.srv_bg_threads,
+			               running_bkg_wrk, wrk_queue);
+		} else if (strcasecmp(type, CMD_STATUS_CONFIG) == 0) {
+			ret = snprintf(buff, sizeof(buff), "%s", configure_summary);
+		} else if (strcasecmp(type, CMD_STATUS_CERT) == 0) {
+			uint8_t pin[128];
+			size_t pin_len = server_cert_pin(args->server, pin, sizeof(pin));
+			if (pin_len > 0) {
+				ret = snprintf(buff, sizeof(buff), "%.*s", (int)pin_len, pin);
+			} else {
+				ret = snprintf(buff, sizeof(buff), STATUS_EMPTY);
+			}
+		} else {
+			return KNOT_EINVAL;
+		}
+		if (ret <= 0 || ret >= sizeof(buff)) {
+			return KNOT_ESPACE;
+		}
 
-	args->data[KNOT_CTL_IDX_DATA] = buff;
+		args->data[KNOT_CTL_IDX_DATA] = buff;
+	}
 
 	return knot_ctl_send(args->ctl, KNOT_CTL_TYPE_DATA, &args->data);
 }
@@ -2187,8 +2146,7 @@ static int ctl_conf_txn(ctl_args_t *args, ctl_cmd_t cmd)
 		}
 		break;
 	case CTL_CONF_ABORT:
-		conf_io_abort(false);
-		ret = KNOT_EOK;
+		ret = conf_io_abort(false);
 		break;
 	case CTL_CONF_COMMIT:
 		// First check the database.
@@ -2204,7 +2162,7 @@ static int ctl_conf_txn(ctl_args_t *args, ctl_cmd_t cmd)
 
 		ret = conf_io_commit(false);
 		if (ret != KNOT_EOK) {
-			conf_io_abort(false);
+			(void)conf_io_abort(false);
 			break;
 		}
 
@@ -2375,7 +2333,7 @@ static int ctl_conf_modify(ctl_args_t *args, ctl_cmd_t cmd)
 			ctl_send_error(args, knot_strerror(ret));
 		}
 	} else {
-		conf_io_abort(true);
+		(void)conf_io_abort(true);
 	}
 
 	return ret;
