@@ -1,17 +1,6 @@
-/*  Copyright (C) 2025 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+/*  Copyright (C) CZ.NIC, z.s.p.o. and contributors
+ *  SPDX-License-Identifier: GPL-2.0-or-later
+ *  For more information, see <https://www.knot-dns.cz/>
  */
 
 #pragma once
@@ -31,6 +20,8 @@
 
 #define CTL_FILTER_FLUSH_OUTDIR		"d"
 
+#define CTL_FILTER_STATUS_LOADING	"l"
+
 #define CTL_FILTER_STATUS_ROLE		"r"
 #define CTL_FILTER_STATUS_SERIAL	"s"
 #define CTL_FILTER_STATUS_TRANSACTION	"t"
@@ -70,6 +61,16 @@
 
 #define STATUS_EMPTY			"-"
 
+/*! Optional 'status' command parameters. */
+#define CMD_STATUS_VERSION              "version"
+#define CMD_STATUS_WORKERS              "workers"
+#define CMD_STATUS_CONFIG               "configure"
+#define CMD_STATUS_CERT                 "cert-key"
+
+/*! 'zone-key-rollover' command key types. */
+#define CMD_ROLLOVER_KSK                "ksk"
+#define CMD_ROLLOVER_ZSK                "zsk"
+
 /*! Control commands. */
 typedef enum {
 	CTL_NONE,
