@@ -1,17 +1,6 @@
-/*  Copyright (C) 2024 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+/*  Copyright (C) CZ.NIC, z.s.p.o. and contributors
+ *  SPDX-License-Identifier: GPL-2.0-or-later
+ *  For more information, see <https://www.knot-dns.cz/>
  */
 
 #pragma once
@@ -20,6 +9,7 @@
 
 #include "contrib/time.h"
 #include "libdnssec/key.h"
+#include "libdnssec/keystore.h"
 #include "knot/conf/conf.h"
 
 /*!
@@ -66,6 +56,15 @@ typedef struct {
 	bool is_zsk;
 } knot_kasp_key_t;
 
+typedef struct {
+	dnssec_keystore_t *keystore;
+	const char *name;
+	unsigned backend;
+	bool ksk_only;
+	bool key_label;
+	size_t count;			/*!< Number of keystores configured. */
+} knot_kasp_keystore_t;
+
 /*!
  * Parent for DS checks.
  */
@@ -132,7 +131,6 @@ typedef struct {
 	bool has_dnskey_sync;
 	bool offline_ksk;
 	bool incremental;
-	bool key_label;
 	unsigned unsafe;
 	uint32_t keytag_remain;
 	uint32_t keytag_modulo;
