@@ -1,17 +1,6 @@
-/*  Copyright (C) 2024 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+/*  Copyright (C) CZ.NIC, z.s.p.o. and contributors
+ *  SPDX-License-Identifier: GPL-2.0-or-later
+ *  For more information, see <https://www.knot-dns.cz/>
  */
 
 #pragma once
@@ -53,7 +42,7 @@ typedef struct zone_events {
 	pthread_mutex_t reschedule_lock;//!< Prevent concurrent reschedule() making mess.
 
 	zone_event_type_t type;		//!< Type of running event.
-	bool running;			//!< Some zone event is being run.
+	time_t running;			//!< Some zone event is being run (since this time).
 	pthread_cond_t *run_end;	//!< Notify this one after finishing a job.
 
 	bool frozen;			//!< Terminated, don't schedule new events.
