@@ -1,17 +1,6 @@
-/*  Copyright (C) 2023 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+/*  Copyright (C) CZ.NIC, z.s.p.o. and contributors
+ *  SPDX-License-Identifier: GPL-2.0-or-later
+ *  For more information, see <https://www.knot-dns.cz/>
  */
 
 #include <assert.h>
@@ -25,6 +14,7 @@
 #include "knot/events/handlers.h"
 #include "knot/events/replan.h"
 #include "knot/zone/digest.h"
+#include "knot/zone/redis.h"
 #include "knot/zone/reverse.h"
 #include "knot/zone/serial.h"
 #include "knot/zone/zone-diff.h"
@@ -55,11 +45,18 @@ static bool allowed_xfr(conf_t *conf, const zone_t *zone)
 	return false;
 }
 
+static int upd_add_rem(const knot_rrset_t *rr, bool add, void *ctx)
+{
+	return add ? zone_update_add(ctx, rr) : zone_update_remove(ctx, rr);
+}
+
 int event_load(conf_t *conf, zone_t *zone)
 {
 	zone_update_t up = { 0 };
 	zone_contents_t *journal_conts = NULL, *zf_conts = NULL;
 	bool old_contents_exist = (zone->contents != NULL), zone_in_journal_exists = false;
+	const char *zone_src = "zone file";
+	struct redisContext *db_ctx = NULL;
 
 	conf_val_t val = conf_zone_get(conf, C_JOURNAL_CONTENT, zone->name);
 	unsigned load_from = conf_opt(&val);
@@ -67,15 +64,22 @@ int event_load(conf_t *conf, zone_t *zone)
 	val = conf_zone_get(conf, C_ZONEFILE_LOAD, zone->name);
 	unsigned zf_from = conf_opt(&val);
 
-	// Note: zone->reverse_from!=NULL almost works, but we need to check if configured even when failed.
-	if (conf_zone_get(conf, C_REVERSE_GEN, zone->name).code == KNOT_EOK ||
-	    zone->cat_members != NULL) { // This should be equivalent to setting catalog-role:generate.
+	bool includes_configured = zone_includes_configured(conf, zone);
+
+	if (includes_configured || zone->cat_members != NULL) { // The latter should be equivalent to setting catalog-role:generate.
 		zf_from = ZONEFILE_LOAD_DIFSE;
 		load_from = JOURNAL_CONTENT_ALL;
 	}
 
 	int ret = KNOT_EOK;
 
+	val = conf_zone_get(conf, C_ZONEFILE_SKIP, zone->name);
+	zone_skip_t skip = { 0 };
+	ret = zone_skip_from_conf(&skip, &val);
+	if (ret != KNOT_EOK) {
+		goto cleanup;
+	}
+
 	// If configured, load journal contents.
 	if (!old_contents_exist && (load_from == JOURNAL_CONTENT_ALL && zf_from != ZONEFILE_LOAD_WHOLE)) {
 		ret = zone_load_from_journal(conf, zone, &journal_conts);
@@ -101,8 +105,76 @@ int event_load(conf_t *conf, zone_t *zone)
 	unsigned digest_alg = conf_opt(&val);
 	bool update_zonemd = (digest_alg != ZONE_DIGEST_NONE);
 
+	uint8_t db_instance = 0;
+	bool db_enabled = conf_zone_rdb_enabled(conf, zone->name, true, &db_instance);
+	if (db_enabled) {
+		zone_src = "database";
+		db_ctx = zone_redis_connect(conf);
+	}
+
+	// Attempt to load changes from database. If fails, load full zone from there later.
+	if (db_enabled && (old_contents_exist || journal_conts != NULL) &&
+	    zone->cat_members == NULL && EMPTY_LIST(zone->include_from) &&
+	    zf_from != ZONEFILE_LOAD_DIFSE && !includes_configured) {
+		zone_redis_err_t err;
+		uint32_t db_serial = 0;
+		ret = zone_redis_serial(db_ctx, db_instance, zone->name, &db_serial, err);
+		if (ret == KNOT_EOK && old_contents_exist && db_serial == zone_contents_serial(zone->contents)) {
+			log_zone_info(zone->name, "database is up-to-date, serial %u", db_serial);
+			goto cleanup;
+		} else if (ret == KNOT_EOK && journal_conts != NULL && db_serial == zone_contents_serial(journal_conts)) {
+			log_zone_info(zone->name, "database is up-to-date with zone-in-journal, serial %u", db_serial);
+			assert(!old_contents_exist);
+			db_enabled = false; // skip both zone_redis_load_upd() and zone_redis_load(), just load from journal. Also skip zone_update_semcheck() later as we do not in fact load from DB.
+		} else if (ret != KNOT_EOK) {
+			log_zone_error(zone->name, "failed to get database status (%s)",
+			               ret == KNOT_ERDB ? err : knot_strerror(ret));
+			goto cleanup; // NOTE this includes the case of KNOT_ENOENT, where DB load is configured but not available
+		}
+
+		if (old_contents_exist) {
+			ret = zone_update_init(&up, zone, UPDATE_INCREMENTAL);
+		} else {
+			ret = zone_update_from_contents(&up, zone, journal_conts, UPDATE_HYBRID);
+		}
+		if (ret != KNOT_EOK) {
+			log_zone_error(zone->name, "failed to initialize update (%s)", knot_strerror(ret));
+			goto cleanup;
+		}
+		if (db_enabled) {
+			uint32_t serial_current = zone_contents_serial(up.new_cont);
+			ret = zone_redis_load_upd(db_ctx, db_instance, zone->name, serial_current,
+			                          upd_add_rem, &up, err);
+			if (ret == KNOT_EOK) {
+				log_zone_info(zone->name, "database updates loaded, instance %u, serial %u -> %u",
+				              db_instance, serial_current, zone_contents_serial(up.new_cont));
+			}
+		}
+		if (ret == KNOT_EOK) {
+			goto load_end; // all OK, skip zone_redis_load() and proceed with incremental zone_update
+		} else if (ret == KNOT_ERDB) {
+			log_zone_error(zone->name, "failed to load updates from database (%s)", err);
+			goto cleanup; // Redis error, surrender
+		} else {
+			zone_update_clear(&up);
+			ret = KNOT_EOK; // just unable to apply DB changesets atop running zone version, go ahead with full zone load from DB
+		}
+	}
+
 	// If configured, attempt to load zonefile.
-	if (zf_from != ZONEFILE_LOAD_NONE && zone->cat_members == NULL) {
+	if ((zf_from != ZONEFILE_LOAD_NONE || db_enabled) && zone->cat_members == NULL) {
+		if (db_enabled) {
+			zone_redis_err_t err;
+			ret = zone_redis_load(db_ctx, db_instance, zone->name, &zf_conts, err);
+			if (ret != KNOT_EOK) {
+				log_zone_error(zone->name, "failed to load from database (%s)",
+				               ret == KNOT_ERDB ? err : knot_strerror(ret));
+				goto cleanup;
+			}
+			zone->zonefile.serial = zone_contents_serial(zf_conts); // for logging
+			goto zonefile_loaded;
+		}
+
 		struct timespec mtime;
 		char *filename = conf_zonefile(conf, zone->name);
 		ret = zonefile_exists(filename, &mtime);
@@ -139,17 +211,13 @@ int event_load(conf_t *conf, zone_t *zone)
 		zone->zonefile.exists = (zf_conts != NULL);
 		zone->zonefile.mtime = mtime;
 
+zonefile_loaded:
 		// If configured, add reverse records to zone contents
-		const knot_dname_t *fail_fwd = NULL;
-		ret = zones_reverse(&zone->reverse_from, zf_conts, &fail_fwd);
-		if (ret == KNOT_ETRYAGAIN) {
-			knot_dname_txt_storage_t forw_str;
-			(void)knot_dname_to_str(forw_str, fail_fwd, sizeof(forw_str));
-			log_zone_warning(zone->name, "waiting for source forward zone '%s'", forw_str);
-			goto cleanup;
-		} else if (ret != KNOT_EOK) {
-			log_zone_error(zone->name, "failed to generate reverse records");
-			goto cleanup;
+		if (includes_configured) {
+			ret = zones_reverse_log(zone, zf_conts);
+			if (ret != KNOT_EOK) {
+				goto cleanup;
+			}
 		}
 
 		// If configured and possible, fix the SOA serial of zonefile.
@@ -158,12 +226,14 @@ int event_load(conf_t *conf, zone_t *zone)
 			uint32_t serial = zone_contents_serial(relevant);
 			uint32_t set = serial_next(serial, conf, zone->name, SERIAL_POLICY_AUTO, 1);
 			zone_contents_set_soa_serial(zf_conts, set);
-			log_zone_info(zone->name, "zone file parsed, serial updated %u -> %u",
-			              zone->zonefile.serial, set);
+			log_zone_info(zone->name, "%s loaded%s%.0u, serial updated %u -> %u",
+			              zone_src, (db_enabled ? ", instance " : ""),
+			              db_instance, zone->zonefile.serial, set);
 			zone->zonefile.serial = set;
 		} else {
-			log_zone_info(zone->name, "zone file parsed, serial %u",
-			              zone->zonefile.serial);
+			log_zone_info(zone->name, "%s loaded%s%.0u, serial %u",
+			              zone_src, (db_enabled ? ", instance " : ""),
+			              db_instance, zone->zonefile.serial);
 		}
 
 		// If configured and appliable to zonefile, load journal changes.
@@ -199,7 +269,15 @@ int event_load(conf_t *conf, zone_t *zone)
 		}
 	}
 
-	bool ignore_dnssec = (do_diff && dnssec_enable);
+	if (do_diff && dnssec_enable) {
+		ret = zone_skip_add_dnssec_diff(&skip);
+	}
+	if (update_zonemd && ret == KNOT_EOK) {
+		ret = zone_skip_add(&skip, "zonemd");
+	}
+	if (ret != KNOT_EOK) {
+		goto cleanup;
+	}
 
 	// Create zone_update structure according to current state.
 	if (old_contents_exist) {
@@ -224,8 +302,7 @@ int event_load(conf_t *conf, zone_t *zone)
 			zu_from_zf_conts = true;
 		} else {
 			// compute ZF diff and if success, apply it
-			ret = zone_update_from_differences(&up, zone, NULL, zf_conts, UPDATE_INCREMENTAL,
-			                                   ignore_dnssec, update_zonemd);
+			ret = zone_update_from_differences(&up, zone, NULL, zf_conts, UPDATE_INCREMENTAL, &skip);
 		}
 	} else {
 		if (journal_conts != NULL && zf_from != ZONEFILE_LOAD_WHOLE) {
@@ -235,7 +312,7 @@ int event_load(conf_t *conf, zone_t *zone)
 			} else {
 				// load zone-in-journal, compute ZF diff and if success, apply it
 				ret = zone_update_from_differences(&up, zone, journal_conts, zf_conts,
-				                                   UPDATE_HYBRID, ignore_dnssec, update_zonemd);
+				                                   UPDATE_HYBRID, &skip);
 				if (ret == KNOT_ESEMCHECK || ret == KNOT_ERANGE) {
 					log_zone_warning(zone->name,
 					                 "zone file changed with SOA serial %s, "
@@ -273,13 +350,13 @@ load_end:
 			}
 			break;
 		case KNOT_ESEMCHECK:
-			log_zone_warning(zone->name, "zone file changed without SOA serial update");
+			log_zone_warning(zone->name, "%s changed without SOA serial update", zone_src);
 			break;
 		case KNOT_ERANGE:
 			if (serial_compare(zone->zonefile.serial, zone_contents_serial(zone->contents)) == SERIAL_INCOMPARABLE) {
-				log_zone_warning(zone->name, "zone file changed with incomparable SOA serial");
+				log_zone_warning(zone->name, "%s changed with incomparable SOA serial", zone_src);
 			} else {
-				log_zone_warning(zone->name, "zone file changed with decreased SOA serial");
+				log_zone_warning(zone->name, "%s changed with decreased SOA serial", zone_src);
 			}
 			break;
 		}
@@ -292,6 +369,13 @@ load_end:
 	zf_conts = NULL;
 	journal_conts = NULL;
 
+	if (db_enabled) {
+		ret = zone_update_semcheck(conf, &up);
+		if (ret != KNOT_EOK) {
+			goto cleanup;
+		}
+	}
+
 	ret = zone_update_verify_digest(conf, &up);
 	if (ret != KNOT_EOK) {
 		goto cleanup;
@@ -318,7 +402,7 @@ load_end:
 			log_zone_warning(zone->name,
 			                 "with automatic DNSSEC signing and outgoing transfers enabled, "
 			                 "'zonefile-load: difference' should be set to avoid malformed "
-			                 "IXFR after manual zone file update");
+			                 "IXFR after manual %s update", zone_src);
 		}
 	} else if (update_zonemd) {
 		/* Don't update ZONEMD if no change and ZONEMD is up-to-date.
@@ -380,6 +464,13 @@ load_end:
 		(void)snprintf(new_serial_str, sizeof(new_serial_str), " -> %u", new_serial);
 	}
 
+	/* Heursitsics: this catches most of the cases when the zone load doesn'T
+	 * introduce any change to zone contents and external validation is useless.
+	 */
+	if (!zone_journal_same_serial(up.zone, new_serial)) {
+		up.flags |= UPDATE_EVREQ;
+	}
+
 	// Commit zone_update back to zone (including journal update, rcu,...).
 	ret = zone_update_commit(conf, &up);
 	if (ret != KNOT_EOK) {
@@ -415,6 +506,9 @@ load_end:
 	if (!zone_timers_serial_notified(&zone->timers, new_serial)) {
 		zone_schedule_notify(conf, zone, 0);
 	}
+	zone_redis_disconnect(db_ctx, true);
+	zone_skip_free(&skip);
+	zone->started = true;
 
 	return KNOT_EOK;
 
@@ -425,6 +519,9 @@ cleanup:
 	zone_update_clear(&up);
 	zone_contents_deep_free(zf_conts);
 	zone_contents_deep_free(journal_conts);
+	zone_redis_disconnect(db_ctx, true);
+	zone_skip_free(&skip);
+	zone->started = true;
 
 	return (dontcare_load_error(conf, zone) ? KNOT_EOK : ret);
 }
