@@ -1,17 +1,6 @@
-/*  Copyright (C) 2023 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+/*  Copyright (C) CZ.NIC, z.s.p.o. and contributors
+ *  SPDX-License-Identifier: GPL-2.0-or-later
+ *  For more information, see <https://www.knot-dns.cz/>
  */
 
 #include <stdio.h>
@@ -602,7 +591,7 @@ static int geo_conf_yparse(check_ctx_t *check, geoip_ctx_t *ctx)
 		ret = KNOT_ENOMEM;
 		goto cleanup;
 	}
-	if (zs_init(scanner, NULL, KNOT_CLASS_IN, ctx->ttl) != 0) {
+	if (zs_init(scanner, ".", KNOT_CLASS_IN, ctx->ttl) != 0) {
 		ret = KNOT_EPARSEFAIL;
 		goto cleanup;
 	}
