@@ -1,17 +1,6 @@
-/*  Copyright (C) 2024 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+/*  Copyright (C) CZ.NIC, z.s.p.o. and contributors
+ *  SPDX-License-Identifier: GPL-2.0-or-later
+ *  For more information, see <https://www.knot-dns.cz/>
  */
 
 #include <urcu.h>
@@ -106,7 +95,7 @@ static void axfr_query_cleanup(knotd_qdata_t *qdata)
 	mm_free(qdata->mm, axfr);
 
 	/* Allow zone changes (finished). */
-	rcu_read_unlock();
+	pthread_rwlock_unlock(&qdata->extra->contents->xfrout_lock);
 }
 
 static void axfr_answer_finished(knotd_qdata_t *qdata, knot_pkt_t *pkt, int state)
@@ -182,7 +171,7 @@ static int axfr_query_init(knotd_qdata_t *qdata)
 	qdata->extra->ext_finished = &axfr_answer_finished;
 
 	/* No zone changes during multipacket answer (unlocked in axfr_answer_cleanup) */
-	rcu_read_lock();
+	pthread_rwlock_rdlock(&qdata->extra->contents->xfrout_lock);
 
 	return KNOT_EOK;
 }
