@@ -1,17 +1,6 @@
-/*  Copyright (C) 2024 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+/*  Copyright (C) CZ.NIC, z.s.p.o. and contributors
+ *  SPDX-License-Identifier: GPL-2.0-or-later
+ *  For more information, see <https://www.knot-dns.cz/>
  */
 
 #pragma once
@@ -19,6 +8,7 @@
 #include <stdint.h>
 #include <sys/socket.h>
 
+#include "knot/conf/conf.h"
 #include "libknot/quic/tls_common.h"
 
 struct knot_request;
@@ -36,19 +26,25 @@ typedef struct knot_tls_req_ctx {
  *
  * \param ctx               Context structure to be initialized.
  * \param fd                Opened TCP connection file descriptor.
+ * \param remote            Remote address for purposes of TLS session resumption.
+ * \param local             Local address for purposes of TLS session resumption.
  * \param local_creds       Local TLS credentials.
- * \param peer_pin          TLS peer pin.
- * \param peer_pin_len      TLS peer pin length.
+ * \param peer_hostnames    Permittable TLS peer hostnames.
+ * \param peer_pins         Permittable TLS peer PINs.
+ * \param reused_fd[out]    Indicates successful TLS session resumption.
  * \param io_timeout_ms     Configured io-timeout for TLS connection.
  *
  * \return KNOT_E*
  */
-int knot_tls_req_ctx_init(knot_tls_req_ctx_t *ctx, int fd,
+int knot_tls_req_ctx_init(knot_tls_req_ctx_t *ctx,
+                          int fd,
                           const struct sockaddr_storage *remote,
                           const struct sockaddr_storage *local,
                           const struct knot_creds *local_creds,
-                          const uint8_t *peer_pin, uint8_t peer_pin_len,
-                          bool *reused_fd, int io_timeout_ms);
+                          const char *const peer_hostnames[RMT_MAX_PINS],
+                          const uint8_t *const peer_pins[RMT_MAX_PINS],
+                          bool *reused_fd,
+                          int io_timeout_ms);
 
 /*!
  * \brief Maintain the TLS requestor context (update session ticket).
