@@ -1,17 +1,6 @@
-/*  Copyright (C) 2025 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+/*  Copyright (C) CZ.NIC, z.s.p.o. and contributors
+ *  SPDX-License-Identifier: GPL-2.0-or-later
+ *  For more information, see <https://www.knot-dns.cz/>
  */
 
 #ifdef __APPLE__
@@ -20,6 +9,7 @@
 
 #include <assert.h>
 #include <gnutls/x509.h>
+#include <string.h>
 #include <sys/types.h>   // OpenBSD
 #include <netinet/tcp.h> // TCP_FASTOPEN
 #include <sys/resource.h>
@@ -28,9 +18,6 @@
 #include "libknot/yparser/ypschema.h"
 #include "libknot/xdp.h"
 #include "libknot/quic/tls.h"
-#ifdef ENABLE_QUIC
-#include "libknot/quic/quic.h" // knot_quic_session_*
-#endif // ENABLE_QUIC
 #include "knot/common/log.h"
 #include "knot/common/stats.h"
 #include "knot/common/systemd.h"
@@ -44,7 +31,9 @@
 #include "knot/server/udp-handler.h"
 #include "knot/server/tcp-handler.h"
 #include "knot/updates/acl.h"
+#include "knot/zone/redis.h"
 #include "knot/zone/timers.h"
+#include "knot/zone/zone.h"
 #include "knot/zone/zonedb-load.h"
 #include "knot/worker/pool.h"
 #include "contrib/base64.h"
@@ -65,6 +54,7 @@
 #endif
 
 #define SESSION_TICKET_POOL_TIMEOUT 1200
+#define REDIS_CONN_POOL_TIMEOUT (4 * 60)
 
 #define QUIC_LOG "QUIC/TLS, "
 
@@ -562,7 +552,7 @@ static void log_sock_conf(conf_t *conf)
 	}
 }
 
-static int check_file(char *path, char *role)
+static int check_file(const char *path, char *role)
 {
 	if (path == NULL) {
 		return KNOT_EOK;
@@ -590,6 +580,10 @@ static int init_creds(conf_t *conf, server_t *server)
 {
 	char *cert_file = conf_tls(conf, C_CERT_FILE);
 	char *key_file = conf_tls(conf, C_KEY_FILE);
+	conf_val_t cafiles_val = conf_get(conf, C_SERVER, C_CA_FILE);
+	size_t nfiles = conf_val_count(&cafiles_val);
+	const char *ca_files[nfiles + 1];
+	bool system_ca = false;
 
 	int ret = check_file(cert_file, "certificate");
 	if (ret != KNOT_EOK) {
@@ -626,16 +620,32 @@ static int init_creds(conf_t *conf, server_t *server)
 		goto failed;
 	}
 
+	memset(ca_files, 0, sizeof(ca_files));
+	for (size_t i = 0; cafiles_val.code == KNOT_EOK; conf_val_next(&cafiles_val)) {
+		const char *file = conf_str(&cafiles_val);
+		if (*file == '\0') {
+			system_ca = true;
+		} else if ((ret = check_file(file, "ca")) != KNOT_EOK) {
+			goto failed;
+		} else {
+			ca_files[i++] = file;
+		}
+	}
+
 	if (server->quic_creds == NULL) {
-		server->quic_creds = knot_creds_init(key_file, cert_file, uid, gid);
-		if (server->quic_creds == NULL) {
-			log_error(QUIC_LOG "failed to initialize server credentials");
-			ret = KNOT_ERROR;
+		ret = knot_creds_init(&server->quic_creds, key_file, cert_file, ca_files,
+		                      system_ca, uid, gid);
+		if (ret != KNOT_EOK) {
+			log_error(QUIC_LOG "failed to initialize credentials or to load certificates (%s)",
+			          knot_strerror(ret));
 			goto failed;
 		}
 	} else {
-		ret = knot_creds_update(server->quic_creds, key_file, cert_file, uid, gid);
+		ret = knot_creds_update(server->quic_creds, key_file, cert_file, ca_files,
+		                        system_ca, uid, gid);
 		if (ret != KNOT_EOK) {
+			log_error(QUIC_LOG "failed to initialize credentials or to load certificates (%s)",
+			          knot_strerror(ret));
 			goto failed;
 		}
 	}
@@ -825,16 +835,161 @@ static int configure_sockets(conf_t *conf, server_t *s)
 	return KNOT_EOK;
 }
 
+#ifdef ENABLE_REDIS
+#include "contrib/openbsd/strlcpy.h"
+#include "contrib/strtonum.h"
+#include "redis/knot.h"
+#include "knot/common/hiredis.h"
+
+#define RDB_TIMESTAMP_SIZE 42 // 2x uint64_t as string (2x20) + dash separator + zero byte.
+
+static void rdb_process_event(redisReply *reply, knot_zonedb_t *zone_db,
+                              char since[static RDB_TIMESTAMP_SIZE])
+{
+	redisReply *timestamp = reply->element[0];
+	redisReply *data = reply->element[1];
+	if (data->type != REDIS_REPLY_ARRAY ||
+	    data->elements % 2 != 0 ||
+	    timestamp->len > RDB_TIMESTAMP_SIZE) {
+		goto failed;
+	}
+	strlcpy(since, timestamp->str, RDB_TIMESTAMP_SIZE);
+
+	uint8_t type = 0;
+	uint8_t instance = 0;
+	uint32_t serial = 0;
+	const knot_dname_t *origin = NULL;
+
+	for (int idx = 0; idx < data->elements; ++idx) {
+		redisReply *key = data->element[idx++];
+		redisReply *val = data->element[idx];
+		if (key->type != REDIS_REPLY_STRING) {
+			goto failed;
+		}
+		if (strcmp(key->str, RDB_EVENT_ARG_EVENT) == 0) {
+			if (str_to_u8(val->str, &type) != KNOT_EOK) {
+				goto failed;
+			}
+		} else if (strcmp(key->str, RDB_EVENT_ARG_ORIGIN) == 0) {
+			origin = (const knot_dname_t *)val->str;
+		} else if (strcmp(key->str, RDB_EVENT_ARG_INSTANCE) == 0) {
+			if (str_to_u8(val->str, &instance) != KNOT_EOK) {
+				goto failed;
+			}
+		} else if (strcmp(key->str, RDB_EVENT_ARG_SERIAL) == 0) {
+			if (str_to_u32(val->str, &serial) != KNOT_EOK) {
+				goto failed;
+			}
+		}
+	}
+
+	if (type == 0 || origin == NULL || instance == 0) {
+		goto failed;
+	}
+
+	uint8_t db_instance = 0;
+	bool db_enabled = conf_zone_rdb_enabled(conf(), origin, true, &db_instance);
+	if (!db_enabled || db_instance != instance) {
+		return;
+	}
+	zone_t *zone = knot_zonedb_find(zone_db, origin);
+	if (zone == NULL) {
+		return;
+	}
+
+	switch (type) {
+	case RDB_EVENT_ZONE:
+	case RDB_EVENT_UPD:
+		zone_schedule_update(conf(), zone, ZONE_EVENT_LOAD);
+		break;
+	default:
+		break;
+	}
+
+	return;
+failed:
+	log_error("rdb, invalid event parameters");
+}
+
+static void rdb_process_events(redisReply *reply, knot_zonedb_t *zone_db,
+                               char since[static RDB_TIMESTAMP_SIZE])
+{
+	if (reply->type != REDIS_REPLY_ARRAY) {
+		log_error("rdb, unexpected response");
+		return;
+	}
+
+	for (int idx = 0; idx < reply->elements; ++idx) {
+		if (reply->element[idx]->type != REDIS_REPLY_ARRAY) {
+			log_error("rdb, unexpected response");
+			continue;
+		}
+		redisReply *events = reply->element[idx]->element[1];
+		for (int event_idx = 0; event_idx < events->elements; ++event_idx) {
+			rdb_process_event(events->element[event_idx], zone_db, since);
+		}
+	}
+}
+
+static int rdb_listener_run(struct dthread *thread)
+{
+	server_t *s = thread->data;
+
+	s->rdb_ctx = NULL;
+	char since[RDB_TIMESTAMP_SIZE] = "$";
+
+	while (thread->state & ThreadActive) {
+		if (s->rdb_ctx == NULL) {
+			s->rdb_ctx = rdb_connect(conf());
+			if (s->rdb_ctx == NULL) {
+				log_error("rdb, failed to connect");
+				sleep(2);
+				continue;
+			} else if (!rdb_compatible(s->rdb_ctx)) {
+				log_error("rdb, incompatible CPU endianness");
+				break;
+			}
+		}
+
+		redisReply *reply = redisCommand(s->rdb_ctx, "XREAD BLOCK %d STREAMS %b %s",
+		                                 10000, RDB_EVENT_KEY, strlen(RDB_EVENT_KEY), since);
+		if (reply == NULL) {
+			if (thread->state & ThreadDead) {
+				break;
+			}
+			if (s->rdb_ctx->err != REDIS_OK) {
+				log_error("rdb, failed to read events (%s)", s->rdb_ctx->errstr);
+			}
+			sleep(2);
+			continue;
+		}
+		if (reply->type == REDIS_REPLY_NIL) {
+			freeReplyObject(reply);
+			continue;
+		}
+
+		knot_zonedb_t *zone_db = s->zone_db;
+		if (zone_db != NULL) {
+			rdb_process_events(reply, zone_db, since);
+		}
+		freeReplyObject(reply);
+	}
+
+	rdb_disconnect(s->rdb_ctx, false);
+	s->rdb_ctx = NULL;
+
+	return KNOT_EOK;
+}
+#endif // ENABLE_REDIS
+
 int server_init(server_t *server, int bg_workers)
 {
 	if (server == NULL) {
 		return KNOT_EINVAL;
 	}
 
-	/* Clear the structure. */
 	memset(server, 0, sizeof(server_t));
 
-	/* Initialize event scheduler. */
 	if (evsched_init(&server->sched, server) != KNOT_EOK) {
 		return KNOT_ENOMEM;
 	}
@@ -845,12 +1000,27 @@ int server_init(server_t *server, int bg_workers)
 		return KNOT_ENOMEM;
 	}
 
+#ifdef ENABLE_REDIS
+	conf_val_t rdb = conf_db_param(conf(), C_ZONE_DB_LISTEN);
+	if (rdb.code == KNOT_EOK) {
+		server->rdb_events = dt_create(1, rdb_listener_run, NULL, server);
+		if (server->rdb_events == NULL) {
+			worker_pool_destroy(server->workers);
+			evsched_deinit(&server->sched);
+			return KNOT_ENOMEM;
+		}
+	}
+#endif // ENABLE_REDIS
+
 	int ret = catalog_update_init(&server->catalog_upd);
 	if (ret != KNOT_EOK) {
+		dt_stop(server->rdb_events);
+		dt_delete(&server->rdb_events);
 		worker_pool_destroy(server->workers);
 		evsched_deinit(&server->sched);
 		return ret;
 	}
+	ATOMIC_INIT(server->catalog_upd_signal, false);
 
 	pthread_rwlock_init(&server->ctl_lock, NULL);
 
@@ -902,6 +1072,9 @@ void server_deinit(server_t *server)
 	/* Free threads and event handlers. */
 	worker_pool_destroy(server->workers);
 
+	/* Free optional zone DB event thread. */
+	dt_delete(&server->rdb_events);
+
 	/* Free zone database. */
 	knot_zonedb_deep_free(&server->zone_db, true);
 
@@ -915,6 +1088,7 @@ void server_deinit(server_t *server)
 	catalog_update_clear(&server->catalog_upd);
 	catalog_update_deinit(&server->catalog_upd);
 	catalog_deinit(&server->catalog);
+	ATOMIC_DEINIT(server->catalog_upd_signal);
 
 	/* Close persistent timers DB. */
 	knot_lmdb_deinit(&server->timerdb);
@@ -930,6 +1104,8 @@ void server_deinit(server_t *server)
 	global_conn_pool = NULL;
 	conn_pool_deinit(global_sessticket_pool);
 	global_sessticket_pool = NULL;
+	conn_pool_deinit(global_redis_pool);
+	global_redis_pool = NULL;
 	knot_unreachables_deinit(&global_unreachables);
 
 	knot_creds_free(server->quic_creds);
@@ -981,42 +1157,19 @@ static void server_free_handler(iohandler_t *h)
 	free(h->thread_id);
 }
 
-static void worker_wait_cb(worker_pool_t *pool)
-{
-	systemd_zone_load_timeout_notify();
-
-	static uint64_t last_ns = 0;
-	struct timespec now = time_now();
-	uint64_t now_ns = 1000000000 * now.tv_sec + now.tv_nsec;
-	/* Too frequent worker_pool_status() call with many zones is expensive. */
-	if (now_ns - last_ns > 1000000000) {
-		int running, queued;
-		worker_pool_status(pool, true, &running, &queued);
-		systemd_tasks_status_notify(running + queued);
-		last_ns = now_ns;
-	}
-}
-
-int server_start(server_t *server, bool async)
+int server_start_answering(server_t *server)
 {
 	if (server == NULL) {
 		return KNOT_EINVAL;
 	}
 
-	/* Start workers. */
-	worker_pool_start(server->workers);
-
-	/* Wait for enqueued events if not asynchronous. */
-	if (!async) {
-		worker_pool_wait_cb(server->workers, worker_wait_cb);
-		systemd_tasks_status_notify(0);
+	assert(server->state & ServerRunning);
+	if (server->state & ServerAnswering) {
+		return KNOT_EOK;
 	}
 
-	/* Start evsched handler. */
-	evsched_start(&server->sched);
-
-	/* Start I/O handlers. */
-	server->state |= ServerRunning;
+	log_info("server started");
+	server->state |= ServerAnswering;
 	for (int proto = IO_UDP; proto <= IO_XDP; ++proto) {
 		if (server->handlers[proto].size > 0) {
 			int ret = dt_start(server->handlers[proto].handler.unit);
@@ -1029,12 +1182,38 @@ int server_start(server_t *server, bool async)
 	return KNOT_EOK;
 }
 
+int server_start(server_t *server, bool answering)
+{
+	if (server == NULL) {
+		return KNOT_EINVAL;
+	}
+
+	/* Start workers. */
+	worker_pool_start(server->workers);
+
+	/* Start zone DB event loop. */
+	dt_start(server->rdb_events);
+
+	/* Start evsched handler. */
+	evsched_start(&server->sched);
+
+	server->state |= ServerRunning;
+
+	/* Start I/O handlers. */
+	if (answering) {
+		return server_start_answering(server);
+	}
+
+	return KNOT_EOK;
+}
+
 void server_wait(server_t *server)
 {
 	if (server == NULL) {
 		return;
 	}
 
+	assert(!(server->state & ServerRunning));
 	evsched_join(&server->sched);
 	worker_pool_join(server->workers);
 
@@ -1043,6 +1222,8 @@ void server_wait(server_t *server)
 			server_free_handler(&server->handlers[proto].handler);
 		}
 	}
+
+	server->state &= ~ServerAnswering;
 }
 
 static int reload_conf(conf_t *new_conf)
@@ -1326,11 +1507,16 @@ int server_reload(server_t *server, reload_t mode)
 			return ret;
 		}
 
+		ret = conf_activate_modules(new_conf, server, NULL, new_conf->query_modules,
+		                            &new_conf->query_plan);
+		if (ret != KNOT_EOK) {
+			conf_free(new_conf);
+			systemd_ready_notify();
+			return ret;
+		}
+
 		ATOMIC_SET(server->stats.tcp_io_timeout, 0);
 		ATOMIC_SET(server->stats.tcp_idle_timeout, 0);
-
-		conf_activate_modules(new_conf, server, NULL, new_conf->query_modules,
-		                      &new_conf->query_plan);
 	}
 
 	conf_update_flag_t upd_flags = CONF_UPD_FNOFREE;
@@ -1377,11 +1563,25 @@ int server_reload(server_t *server, reload_t mode)
 
 void server_stop(server_t *server)
 {
+	if (server == NULL || !(server->state & ServerRunning)) {
+		return;
+	}
+
 	log_info("stopping server");
 	systemd_stopping_notify();
 
+#ifdef ENABLE_REDIS
+	/* Interrupt and stop XREAD BLOCK loop. */
+	if (server->rdb_ctx != NULL) {
+		redisSetTimeout(server->rdb_ctx, (struct timeval){ 0, 1 });
+	}
+	dt_stop(server->rdb_events);
+#endif // ENABLE_REDIS
+
 	/* Stop scheduler. */
 	evsched_stop(&server->sched);
+	/* Mark the server is shutting down. */
+	server->state |= ServerShutting;
 	/* Interrupt background workers. */
 	worker_pool_stop(server->workers);
 
@@ -1463,6 +1663,18 @@ static void free_sess_ticket(intptr_t ptr)
 	}
 }
 
+static void free_redis_conn(intptr_t ptr)
+{
+	if (ptr != CONN_POOL_FD_INVALID) {
+		zone_redis_disconnect((void *)ptr, false);
+	}
+}
+
+static bool invalid_redis_conn(intptr_t ptr)
+{
+	return ptr == CONN_POOL_FD_INVALID || !zone_redis_ping((void *)ptr);
+}
+
 static int reconfigure_remote_pool(conf_t *conf, server_t *server)
 {
 	conf_val_t val = conf_get(conf, C_SRV, C_RMT_POOL_LIMIT);
@@ -1493,6 +1705,24 @@ static int reconfigure_remote_pool(conf_t *conf, server_t *server)
 			}
 			global_sessticket_pool = new_pool;
 		}
+	} else if (server->quic_creds != NULL) {
+		static uint64_t hash = 0;
+		uint64_t curr_hash = knot_creds_hash(server->quic_creds);
+		if (hash != curr_hash) {
+			conn_pool_purge(global_sessticket_pool);
+		}
+		hash = curr_hash;
+	}
+
+	val = conf_get(conf, C_DB, C_ZONE_DB_LISTEN);
+	if (global_redis_pool == NULL && val.code == KNOT_EOK) {
+		size_t bg_wrkrs = conf_bg_threads(conf);
+		conn_pool_t *new_pool = conn_pool_init(bg_wrkrs, REDIS_CONN_POOL_TIMEOUT,
+		                                       free_redis_conn, invalid_redis_conn);
+		if (new_pool == NULL) {
+			return KNOT_ENOMEM;
+		}
+		global_redis_pool = new_pool;
 	}
 
 	val = conf_get(conf, C_SRV, C_RMT_RETRY_DELAY);
@@ -1592,6 +1822,7 @@ void server_update_zones(conf_t *conf, server_t *server, reload_t mode)
 	/* Suspend adding events to worker pool queue, wait for queued events. */
 	log_debug("suspending zone events");
 	evsched_pause(&server->sched);
+	server->state |= ServerShutting;
 	worker_pool_wait(server->workers);
 	log_debug("suspended zone events");
 
@@ -1606,6 +1837,7 @@ void server_update_zones(conf_t *conf, server_t *server, reload_t mode)
 	if (server->zone_db) {
 		knot_zonedb_foreach(server->zone_db, zone_events_start);
 	}
+	server->state &= ~ServerShutting;
 	log_debug("resumed zone events");
 }
 
