@@ -1,17 +1,6 @@
-/*  Copyright (C) 2024 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+/*  Copyright (C) CZ.NIC, z.s.p.o. and contributors
+ *  SPDX-License-Identifier: GPL-2.0-or-later
+ *  For more information, see <https://www.knot-dns.cz/>
  */
 
 #include <assert.h>
@@ -603,10 +592,11 @@ int changeset_walk(const changeset_t *changeset, changeset_walk_callback callbac
 	return KNOT_EOK;
 }
 
-void changeset_print(const changeset_t *changeset, FILE *outfile, bool color)
+int changeset_print(const changeset_t *changeset, FILE *outfile, bool color)
 {
 	size_t buflen = 1024;
 	char *buff = malloc(buflen);
+	int ret = KNOT_EOK;
 
 	knot_dump_style_t style = KNOT_DUMP_STYLE_DEFAULT;
 	style.now = knot_time();
@@ -616,20 +606,25 @@ void changeset_print(const changeset_t *changeset, FILE *outfile, bool color)
 		fprintf(outfile, "%s;; Removed%s\n", style.color, COL_RST(color));
 	}
 	if (changeset->soa_from != NULL && buff != NULL) {
-		(void)knot_rrset_txt_dump(changeset->soa_from, &buff, &buflen, &style);
+		ret = knot_rrset_txt_dump(changeset->soa_from, &buff, &buflen, &style);
 		fprintf(outfile, "%s%s%s", style.color, buff, COL_RST(color));
 	}
-	(void)zone_dump_text(changeset->remove, outfile, false, style.color);
+	if (ret >= 0 && changeset->remove != NULL) { // Can be NULL if zone-in-journal
+		ret = zone_dump_text(changeset->remove, NULL, outfile, false, style.color);
+	}
 
 	style.color = COL_GRN(color);
 	if (changeset->soa_to != NULL || !zone_contents_is_empty(changeset->add)) {
 		fprintf(outfile, "%s;; Added%s\n", style.color, COL_RST(color));
 	}
-	if (changeset->soa_to != NULL && buff != NULL) {
-		(void)knot_rrset_txt_dump(changeset->soa_to, &buff, &buflen, &style);
+	if (changeset->soa_to != NULL && buff != NULL && ret >= 0) {
+		ret = knot_rrset_txt_dump(changeset->soa_to, &buff, &buflen, &style);
 		fprintf(outfile, "%s%s%s", style.color, buff, COL_RST(color));
 	}
-	(void)zone_dump_text(changeset->add, outfile, false, style.color);
+	if (ret >= 0) {
+		ret = zone_dump_text(changeset->add, NULL, outfile, false, style.color);
+	}
 
 	free(buff);
+	return ret >= 0 ? KNOT_EOK : ret;
 }
