@@ -1,17 +1,6 @@
-/*  Copyright (C) 2024 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+/*  Copyright (C) CZ.NIC, z.s.p.o. and contributors
+ *  SPDX-License-Identifier: GPL-2.0-or-later
+ *  For more information, see <https://www.knot-dns.cz/>
  */
 
 #include <signal.h>
@@ -27,6 +16,7 @@
 #include "knot/zone/adds_tree.h"
 #include "knot/zone/adjust.h"
 #include "knot/zone/digest.h"
+#include "knot/zone/redis.h"
 #include "knot/zone/serial.h"
 #include "knot/zone/zone-diff.h"
 #include "knot/zone/zonefile.h"
@@ -165,7 +155,7 @@ int zone_update_init(zone_update_t *update, zone_t *zone, zone_update_flags_t fl
 
 int zone_update_from_differences(zone_update_t *update, zone_t *zone, zone_contents_t *old_cont,
                                  zone_contents_t *new_cont, zone_update_flags_t flags,
-                                 bool ignore_dnssec, bool ignore_zonemd)
+                                 zone_skip_t *skip)
 {
 	if (update == NULL || zone == NULL || new_cont == NULL ||
 	    !(flags & (UPDATE_INCREMENTAL | UPDATE_HYBRID)) || (flags & UPDATE_FULL)) {
@@ -188,7 +178,7 @@ int zone_update_from_differences(zone_update_t *update, zone_t *zone, zone_conte
 		old_cont = zone->contents;
 	}
 
-	ret = zone_contents_diff(old_cont, new_cont, &diff, ignore_dnssec, ignore_zonemd);
+	ret = zone_contents_diff(old_cont, new_cont, &diff, skip);
 	switch (ret) {
 	case KNOT_ENODIFF:
 	case KNOT_ESEMCHECK:
@@ -290,7 +280,7 @@ int zone_update_start_extra(zone_update_t *update, conf_t *conf)
 		}
 
 		ret = zone_contents_diff(update->init_cont, update->new_cont,
-		                         &update->extra_ch, false, false);
+		                         &update->extra_ch, NULL);
 		if (ret != KNOT_EOK) {
 			return ret;
 		}
@@ -687,6 +677,11 @@ static int commit_journal(conf_t *conf, zone_update_t *update)
 {
 	conf_val_t val = conf_zone_get(conf, C_JOURNAL_CONTENT, update->zone->name);
 	unsigned content = conf_opt(&val);
+
+	if (zone_includes_configured(conf, update->zone)){
+		content = JOURNAL_CONTENT_ALL;
+	}
+
 	int ret = KNOT_EOK;
 	if (update->flags & UPDATE_NO_CHSET) {
 		zone_diff_t diff;
@@ -710,6 +705,75 @@ static int commit_journal(conf_t *conf, zone_update_t *update)
 	return ret;
 }
 
+static int redis_wr_rr(const knot_rrset_t *rr, void *ctx)
+{
+	return zone_redis_write_rrset(ctx, rr);
+}
+
+static int redis_wr_node(zone_node_t *node, void *ctx)
+{
+	return zone_redis_write_node(ctx, node);
+}
+
+static int commit_redis(conf_t *conf, zone_update_t *update)
+{
+	uint8_t db_instance = 0;
+	bool db_enabled = conf_zone_rdb_enabled(conf, update->zone->name, false, &db_instance);
+	if (!db_enabled) {
+		return KNOT_EOK;
+	}
+
+	struct redisContext *db_ctx = zone_redis_connect(conf);
+	if (db_ctx == NULL) {
+		return KNOT_ECONN;
+	}
+
+	bool incremental = ((update->flags & (UPDATE_INCREMENTAL | UPDATE_HYBRID)) && update->zone->contents != NULL);
+	if (incremental) {
+		zone_redis_err_t err;
+		uint32_t redis_soa = 0;
+		int soa_ret = zone_redis_serial(db_ctx, db_instance, update->zone->name, &redis_soa, err);
+		incremental = (soa_ret == KNOT_EOK && redis_soa == zone_contents_serial(update->zone->contents));
+	}
+
+	zone_redis_txn_t txn;
+	int ret = zone_redis_txn_begin(&txn, db_ctx, db_instance, update->zone->name, incremental);
+	if (ret != KNOT_EOK) {
+		zone_redis_disconnect(db_ctx, true);
+		return ret;
+	}
+
+	if (incremental) {
+		txn.removals = true;
+		ret = zone_update_foreach(update, false, redis_wr_rr, &txn);
+		if (ret == KNOT_EOK) {
+			txn.removals = false;
+			ret = zone_update_foreach(update, true, redis_wr_rr, &txn);
+		}
+	} else {
+		ret = zone_contents_apply(update->new_cont, redis_wr_node, &txn);
+		if (ret == KNOT_EOK) {
+			ret = zone_contents_nsec3_apply(update->new_cont, redis_wr_node, &txn);
+		}
+	}
+
+	if (ret == KNOT_EOK) {
+		ret = zone_redis_txn_commit(&txn);
+	}
+	if (ret != KNOT_EOK) {
+		if (ret == KNOT_ERDB) {
+			log_zone_error(update->zone->name, "rdb, update aborted (%s)", txn.err);
+		}
+		(void)zone_redis_txn_abort(&txn);
+	} else {
+		log_zone_info(update->zone->name, "database updated, instance %u, serial %u",
+		              db_instance, zone_contents_serial(update->new_cont));
+	}
+
+	zone_redis_disconnect(db_ctx, true);
+	return ret;
+}
+
 static int commit_incremental(conf_t *conf, zone_update_t *update)
 {
 	assert(update);
@@ -780,7 +844,7 @@ static int update_catalog(conf_t *conf, zone_update_t *update)
 
 	if (ret == KNOT_EOK) {
 		log_zone_info(update->zone->name, "enqueued %zd catalog updates", upd_count);
-		update->zone->server->catalog_upd_signal = true;
+		ATOMIC_SET(update->zone->server->catalog_upd_signal, true);
 		if (kill(getpid(), SIGUSR1) != 0) {
 			ret = knot_map_errno();
 		}
@@ -818,6 +882,7 @@ static bool counter_reach(counter_reach_t *counter, size_t increment, size_t lim
  */
 typedef struct {
 	struct rcu_head rcuhead;
+	pthread_t *retry_thread;
 
 	zone_contents_t *free_contents;
 	void (*free_method)(zone_contents_t *);
@@ -827,12 +892,27 @@ typedef struct {
 	size_t new_cont_size;
 } update_clear_ctx_t;
 
+static void *update_clear_retry(void *arg);
+
 static void update_clear(struct rcu_head *param)
 {
 	static counter_reach_t counter = { PTHREAD_MUTEX_INITIALIZER, 0 };
 
 	update_clear_ctx_t *ctx = (update_clear_ctx_t *)param;
 
+	if (ctx->free_contents != NULL) {
+		if (pthread_rwlock_trywrlock(&ctx->free_contents->xfrout_lock) != 0) {
+			if (*ctx->retry_thread == 0 &&
+			    pthread_create(ctx->retry_thread, NULL, update_clear_retry, ctx) == 0) {
+				return;
+			}
+			log_zone_debug(ctx->free_contents->apex->owner,
+			               "disposal of old contents blocked by outstanding zone transfer");
+			pthread_rwlock_wrlock(&ctx->free_contents->xfrout_lock);
+		}
+		pthread_rwlock_unlock(&ctx->free_contents->xfrout_lock);
+	}
+
 	ctx->free_method(ctx->free_contents);
 	apply_cleanup(ctx->cleanup_apply);
 	free(ctx->cleanup_apply);
@@ -844,6 +924,12 @@ static void update_clear(struct rcu_head *param)
 	free(ctx);
 }
 
+static void *update_clear_retry(void *arg)
+{
+	update_clear(arg);
+	return NULL;
+}
+
 static void discard_adds_tree(zone_update_t *update)
 {
 	additionals_tree_free(update->new_cont->adds_tree);
@@ -904,6 +990,116 @@ int zone_update_verify_digest(conf_t *conf, zone_update_t *update)
 	return ret;
 }
 
+typedef struct {
+	FILE *f;
+	char *buf;
+	size_t buflen;
+	knot_dump_style_t style;
+} dump_changeset_ctx_t;
+
+static int dump_rrset(const knot_rrset_t *rr, void *ctx_void)
+{
+	dump_changeset_ctx_t *ctx = ctx_void;
+
+	int ret = knot_rrset_txt_dump(rr, &ctx->buf, &ctx->buflen, &ctx->style);
+	if (ret > 0) {
+		ret = (fprintf(ctx->f, "%s", ctx->buf) < 0 ? knot_map_errno() : 1);
+	}
+	return ret < 0 ? ret : KNOT_EOK;
+}
+
+static int dump_changeset_part(zone_update_t *up, bool additions,
+                               const char *fname, const char *mode)
+{
+	dump_changeset_ctx_t ctx;
+	ctx.f = fopen(fname, mode);
+	if (ctx.f == NULL) {
+		return knot_map_errno();
+	}
+	ctx.buflen = 1024;
+	ctx.buf = malloc(ctx.buflen);
+	if (ctx.buf == NULL) {
+		fclose(ctx.f);
+		return KNOT_ENOMEM;
+	}
+	ctx.style = KNOT_DUMP_STYLE_DEFAULT;
+	ctx.style.now = knot_time();
+
+	(void)fprintf(ctx.f, ";; %s records\n", additions ? "Added" : "Removed");
+	int ret = zone_update_foreach(up, additions, dump_rrset, &ctx);
+	fclose(ctx.f);
+	free(ctx.buf);
+	return ret;
+}
+
+int zone_update_external(conf_t *conf, zone_update_t *update, conf_val_t *ev_id)
+{
+	/* First: dump zone/diff files as/if configured. */
+	conf_val_t val = conf_id_get(conf, C_EXTERNAL, C_DUMP_NEW, ev_id);
+	char *f_new = conf_get_filename(conf, update->zone->name, conf_str(&val));
+	val = conf_id_get(conf, C_EXTERNAL, C_DUMP_REM, ev_id);
+	char *f_rem = conf_get_filename(conf, update->zone->name, conf_str(&val));
+	val = conf_id_get(conf, C_EXTERNAL, C_DUMP_ADD, ev_id);
+	char *f_add = conf_get_filename(conf, update->zone->name, conf_str(&val));
+
+	int ret = KNOT_EOK;
+	if (f_new != NULL && ret == KNOT_EOK) {
+		ret = zonefile_write(f_new, update->new_cont, NULL);
+	}
+	if (f_rem != NULL && ret == KNOT_EOK) {
+		ret = dump_changeset_part(update, false, f_rem, "w");
+	}
+	if (f_add != NULL && ret == KNOT_EOK) {
+		const char *mode = "w";
+		if (f_rem != NULL && strcmp(f_rem, f_add) == 0) {
+			mode = "a";
+		}
+		ret = dump_changeset_part(update, true, f_add, mode);
+	}
+	free(f_new);
+	free(f_rem);
+	free(f_add);
+	if (ret != KNOT_EOK) {
+		log_zone_error(update->zone->name,
+		               "failed to dump new zone version or zone diff for external validation (%s)",
+		               knot_strerror(ret));
+	}
+
+	/* Second: wait on semaphore on user's interaction. */
+	pthread_mutex_lock(&update->zone->cu_lock);
+
+	assert(update->zone->control_update == NULL);
+
+	/* Don't start external validation if shutting down. */
+	if (update->zone->server->state & ServerShutting) {
+		pthread_mutex_unlock(&update->zone->cu_lock);
+		return KNOT_EEXTERNAL;
+	}
+
+	update->zone->control_update = update;
+	update->flags |= UPDATE_WFEV;
+	knot_sem_init(&update->external, 0);
+	pthread_mutex_unlock(&update->zone->cu_lock);
+
+	log_zone_notice(update->zone->name, "waiting for external validation");
+
+	if (conf->cache.srv_dbus_event & DBUS_EVENT_EXTERNAL) {
+		dbus_emit_external_verify(update->zone->name);
+	}
+
+	val = conf_id_get(conf, C_EXTERNAL, C_TIMEOUT, ev_id);
+	knot_sem_timedwait(&update->external, conf_int(&val) * 1000);
+
+	pthread_mutex_lock(&update->zone->cu_lock);
+	update->zone->control_update = NULL;
+	pthread_mutex_unlock(&update->zone->cu_lock);
+
+	knot_sem_post(&update->external);
+	knot_sem_destroy(&update->external);
+
+	return (update->flags & UPDATE_EVOK) ? KNOT_EOK : KNOT_EEXTERNAL;
+}
+
 int zone_update_commit(conf_t *conf, zone_update_t *update)
 {
 	if (conf == NULL || update == NULL) {
@@ -966,6 +1162,15 @@ int zone_update_commit(conf_t *conf, zone_update_t *update)
 		}
 	}
 
+	val = conf_zone_get(conf, C_EXTERNAL_VLDT, update->zone->name);
+	if (val.code == KNOT_EOK && (update->flags & UPDATE_EVREQ)) {
+		ret = zone_update_external(conf, update, &val);
+		if (ret != KNOT_EOK) {
+			discard_adds_tree(update);
+			return ret;
+		}
+	}
+
 	ret = update_catalog(conf, update);
 	if (ret != KNOT_EOK) {
 		log_zone_error(update->zone->name, "failed to process catalog zone (%s)", knot_strerror(ret));
@@ -973,6 +1178,13 @@ int zone_update_commit(conf_t *conf, zone_update_t *update)
 		return ret;
 	}
 
+	ret = commit_redis(conf, update);
+	if (ret != KNOT_EOK) {
+		log_zone_error(update->zone->name, "zone database update failed (%s)", knot_strerror(ret));
+		discard_adds_tree(update);
+		return ret;
+	}
+
 	ret = commit_journal(conf, update);
 	if (ret != KNOT_EOK) {
 		log_zone_error(update->zone->name, "journal update failed (%s)", knot_strerror(ret));
@@ -1000,6 +1212,7 @@ int zone_update_commit(conf_t *conf, zone_update_t *update)
 
 	update_clear_ctx_t *clear_ctx = calloc(1, sizeof(*clear_ctx));
 	if (clear_ctx != NULL) {
+		clear_ctx->retry_thread = &update->zone->update_clear_thr;
 		clear_ctx->free_contents = old_contents;
 		clear_ctx->free_method = (
 			(update->flags & (UPDATE_FULL | UPDATE_HYBRID)) ?
@@ -1013,7 +1226,7 @@ int zone_update_commit(conf_t *conf, zone_update_t *update)
 		log_zone_error(update->zone->name, "failed to deallocate unused memory");
 	}
 
-	zone_local_notify(update->zone);
+	zone_local_notify(conf, update->zone);
 
 	/* Sync zonefile immediately if configured. */
 	val = conf_zone_get(conf, C_ZONEFILE_SYNC, update->zone->name);
@@ -1044,9 +1257,88 @@ bool zone_update_no_change(zone_update_t *update)
 	} else if (update->flags & (UPDATE_INCREMENTAL | UPDATE_HYBRID)) {
 		return changeset_empty(&update->change);
 	} else {
-		/* This branch does not make much sense and FULL update will most likely
-		 * be a change every time anyway, just return false. */
-		return false;
+		return zone_contents_is_empty(update->new_cont);
+	}
+}
+
+static int rrset_foreach(zone_node_t *n, bool subtract_counterpart,
+                         int idx, uint16_t rrtype, // two exclusive possibilities how to define which rrset in the node
+                         rrset_cb_t cb, void *ctx)
+{
+	knot_rrset_t rr = (rrtype == KNOT_RRTYPE_ANY) ? node_rrset_at(n, idx) : node_rrset(n, rrtype);
+	knot_rrset_t rrc = subtract_counterpart ? node_rrset(binode_counterpart(n), rr.type) : (knot_rrset_t){ 0 };
+	if (rrtype == KNOT_RRTYPE_ANY && rr.type == KNOT_RRTYPE_SOA) {
+		return KNOT_EOK; // ignore SOA if rrset specified by idx
+	} else if (knot_rrset_empty(&rrc)) {
+		return cb(&rr, ctx);
+	} else if (knot_rdataset_subset(&rr.rrs, &rrc.rrs)) {
+		return KNOT_EOK;
+	} else {
+		knot_rdataset_t rd_copy = { 0 };
+		int ret = knot_rdataset_copy(&rd_copy, &rr.rrs, NULL);
+		if (ret == KNOT_EOK) {
+			ret = knot_rdataset_subtract(&rd_copy, &rrc.rrs, NULL);
+		}
+		if (ret == KNOT_EOK) {
+			rr.rrs = rd_copy;
+			ret = cb(&rr, ctx);
+		}
+		knot_rdataset_clear(&rd_copy, NULL);
+		return ret;
+	}
+}
+
+static int trees_foreach(zone_tree_t *nodes, zone_tree_t *nsec3_nodes, bool subtract_counterparts,
+                         rrset_cb_t cb, void *ctx)
+{
+	zone_tree_it_t it = { 0 };
+	int ret = zone_tree_it_double_begin(nodes, nsec3_nodes, &it);
+	while (!zone_tree_it_finished(&it) && ret == KNOT_EOK) {
+		zone_node_t *n = zone_tree_it_val(&it);
+		for (int i = 0; i < n->rrset_count && ret == KNOT_EOK; i++) {
+			ret = rrset_foreach(n, subtract_counterparts, i, KNOT_RRTYPE_ANY, cb, ctx);
+		}
+		zone_tree_it_next(&it);
+	}
+	zone_tree_it_free(&it);
+	return ret;
+}
+
+int zone_update_foreach(zone_update_t *update, bool additions, rrset_cb_t cb, void *ctx)
+{
+	if (update == NULL) {
+		return KNOT_EINVAL;
+	}
+
+	if (update->flags & UPDATE_NO_CHSET) {
+		zone_diff_t diff;
+		get_zone_diff(&diff, update);
+		if (!additions) {
+			zone_diff_reverse(&diff);
+		}
+
+		int ret = rrset_foreach(diff.apex, true, 0, KNOT_RRTYPE_SOA, cb, ctx);
+		if (ret == KNOT_EOK) {
+			ret = trees_foreach(&diff.nodes, &diff.nsec3s, true, cb, ctx);
+		}
+		return ret;
+	} else if (update->flags & (UPDATE_INCREMENTAL | UPDATE_HYBRID)) {
+		knot_rrset_t *soa = additions ? update->change.soa_to : update->change.soa_from;
+		zone_contents_t *c = additions ? update->change.add : update->change.remove;
+		int ret = (soa == NULL) ? KNOT_EOK : cb(soa, ctx);
+		if (ret == KNOT_EOK) {
+			ret = trees_foreach(c->nodes, c->nsec3_nodes, false, cb, ctx);
+		}
+		return ret;
+	} else if (additions) {
+		knot_rrset_t soa = node_rrset(update->new_cont->apex, KNOT_RRTYPE_SOA);
+		int ret = knot_rrset_empty(&soa) ? KNOT_EOK : cb(&soa, ctx);
+		if (ret == KNOT_EOK) {
+			ret = trees_foreach(update->new_cont->nodes, update->new_cont->nsec3_nodes, false, cb, ctx);
+		}
+		return ret;
+	} else {
+		return KNOT_EOK;
 	}
 }
 
