@@ -1,17 +1,6 @@
-/*  Copyright (C) 2024 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+/*  Copyright (C) CZ.NIC, z.s.p.o. and contributors
+ *  SPDX-License-Identifier: GPL-2.0-or-later
+ *  For more information, see <https://www.knot-dns.cz/>
  */
 
 #pragma once
@@ -20,6 +9,7 @@
 #include "knot/conf/conf.h"
 #include "knot/updates/changesets.h"
 #include "knot/zone/contents.h"
+#include "knot/zone/skip.h"
 #include "knot/zone/zone.h"
 
 typedef struct {
@@ -39,6 +29,7 @@ typedef struct zone_update {
 	changeset_t extra_ch;        /*!< Extra changeset to store just diff btwn zonefile and result. */
 	apply_ctx_t *a_ctx;          /*!< Context for applying changesets. */
 	uint32_t flags;              /*!< Zone update flags. */
+	knot_sem_t external;         /*!< Lock for external validation. */
 	dnssec_validation_hint_t validation_hint;
 } zone_update_t;
 
@@ -60,6 +51,9 @@ typedef enum {
 	UPDATE_CHANGED_NSEC   = 1 << 7, /*!< This incremental update affects NSEC or NSEC3 nodes in zone. */
 	UPDATE_NO_CHSET       = 1 << 8, /*!< Avoid using changeset and serialize to journal from diff of bi-nodes. */
 	UPDATE_SIGNED_FULL    = 1 << 9, /*!< Full (non-incremental) zone sign took place during this update. */
+	UPDATE_EVREQ          = 1 << 10, /*!< Update requires external validation (if configured). */
+	UPDATE_WFEV           = 1 << 11, /*!< Update waiting for external validation. */
+	UPDATE_EVOK           = 1 << 12, /*!< External validation accepted the update. */
 } zone_update_flags_t;
 
 /*!
@@ -85,14 +79,13 @@ int zone_update_init(zone_update_t *update, zone_t *zone, zone_update_flags_t fl
  * \param old_cont The current zone contents the diff will be against. Probably zone->contents.
  * \param new_cont New zone contents. Will be taken over (and later freed) by zone update.
  * \param flags    Flags for update. Must be UPDATE_INCREMENTAL or UPDATE_HYBRID.
- * \param ignore_dnssec Ignore DNSSEC records.
- * \param ignore_zonemd Ignore ZONEMD records.
+ * \param skip     Record types to be skipped during difference computation.
  *
  * \return KNOT_E*
  */
 int zone_update_from_differences(zone_update_t *update, zone_t *zone, zone_contents_t *old_cont,
                                  zone_contents_t *new_cont, zone_update_flags_t flags,
-                                 bool ignore_dnssec, bool ignore_zonemd);
+                                 zone_skip_t *skip);
 
 /*!
  * \brief Inits a zone update based on new zone contents.
@@ -279,6 +272,19 @@ int zone_update_semcheck(conf_t *conf, zone_update_t *update);
  */
 int zone_update_verify_digest(conf_t *conf, zone_update_t *update);
 
+/*!
+ * \brief Wait for external validation.
+ *
+ * \param conf      Configuration.
+ * \param update    Zone update.
+ * \param ev_id     Conf reference to 'external' section.
+ *
+ * \retval KNOT_EEXTERNAL   External validation failed.
+ * \retval KNOT_EOK         External validation succeeded.
+ * \return KNOT_E*
+ */
+int zone_update_external(conf_t *conf, zone_update_t *update, conf_val_t *ev_id);
+
 /*!
  * \brief Commits all changes to the zone, signs it, saves changes to journal.
  *
@@ -296,6 +302,19 @@ int zone_update_commit(conf_t *conf, zone_update_t *update);
  */
 bool zone_update_no_change(zone_update_t *update);
 
+typedef int (*rrset_cb_t)(const knot_rrset_t *, void *);
+/*!
+ * \brief Run callback for every removed/added RRset in this update.
+ *
+ * \param update       Zone update in question.
+ * \param additions    Apply on addition (removals otherwise).
+ * \param cb           Callback to call for each changed RRset.
+ * \param ctx          Arbitrary context for the callback.
+ *
+ * \return KNOT_E* emitted by the callback or error in iteration.
+ */
+int zone_update_foreach(zone_update_t *update, bool additions, rrset_cb_t cb, void *ctx);
+
 /*!
  * \brief Return whether apex DNSKEY, CDNSKEY, or CDS is updated.
  */
