@@ -1,17 +1,6 @@
-/*  Copyright (C) 2024 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+/*  Copyright (C) CZ.NIC, z.s.p.o. and contributors
+ *  SPDX-License-Identifier: GPL-2.0-or-later
+ *  For more information, see <https://www.knot-dns.cz/>
  */
 
 #include <assert.h>
@@ -192,6 +181,8 @@ zone_contents_t *zone_contents_new(const knot_dname_t *apex_name, bool use_binod
 	}
 	contents->apex->flags |= NODE_FLAGS_APEX;
 	contents->max_ttl = UINT32_MAX;
+	ATOMIC_INIT(contents->dnssec_expire, 0);
+	pthread_rwlock_init(&contents->xfrout_lock, NULL);
 
 	return contents;
 
@@ -522,6 +513,9 @@ void zone_contents_free(zone_contents_t *contents)
 	dnssec_nsec3_params_free(&contents->nsec3_params);
 	additionals_tree_free(contents->adds_tree);
 
+	ATOMIC_DEINIT(contents->dnssec_expire);
+	pthread_rwlock_destroy(&contents->xfrout_lock);
+
 	free(contents);
 }
 
@@ -531,6 +525,8 @@ void zone_contents_deep_free(zone_contents_t *contents)
 		return;
 	}
 
+	pthread_rwlock_wrlock(&contents->xfrout_lock);
+
 	if (contents != NULL) {
 		// Delete NSEC3 tree.
 		(void)zone_tree_apply(contents->nsec3_nodes,
@@ -541,6 +537,8 @@ void zone_contents_deep_free(zone_contents_t *contents)
 		                      destroy_node_rrsets_from_tree, NULL);
 	}
 
+	pthread_rwlock_unlock(&contents->xfrout_lock);
+
 	zone_contents_free(contents);
 }
 
