@@ -1,17 +1,6 @@
-/*  Copyright (C) 2024 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+/*  Copyright (C) CZ.NIC, z.s.p.o. and contributors
+ *  SPDX-License-Identifier: GPL-2.0-or-later
+ *  For more information, see <https://www.knot-dns.cz/>
  */
 
 #include <stdlib.h>
@@ -108,7 +97,7 @@ static int flush_journal(conf_t *conf, zone_t *zone, bool allow_empty_zone, bool
 	char *zonefile = conf_zonefile(conf, zone->name);
 
 	/* Synchronize journal. */
-	ret = zonefile_write(zonefile, contents);
+	ret = zonefile_write_skip(zonefile, contents, conf);
 	rcu_read_unlock();
 	if (ret != KNOT_EOK) {
 		log_zone_warning(zone->name, "failed to update zone file (%s)",
@@ -192,9 +181,11 @@ zone_t* zone_new(const knot_dname_t *name)
 	// Initialize query modules list.
 	init_list(&zone->query_modules);
 
-	init_list(&zone->reverse_from);
+	init_list(&zone->include_from);
 	init_list(&zone->internal_notify);
 
+	ATOMIC_INIT(zone->backup_ctx, NULL);
+
 	return zone;
 }
 
@@ -219,15 +210,27 @@ void zone_free(zone_t **zone_ptr)
 
 	zone_events_deinit(zone);
 
+	if (zone->update_clear_thr) {
+		pthread_join(zone->update_clear_thr, NULL);
+	}
+
+	/* Even when events are done, delayed RCU cleanup might still be ongoing. */
+	knot_sem_wait(&zone->cow_lock);
+
+	/* Free zone contents. Possible wait for XFRout lock. */
+	zone_contents_deep_free(zone->contents);
+
 	knot_dname_free(zone->name, NULL);
 
 	free_ddns_queue(zone);
 	pthread_mutex_destroy(&zone->ddns_lock);
 
 	pthread_mutex_destroy(&zone->cu_lock);
+	knot_sem_post(&zone->cow_lock);
 	knot_sem_destroy(&zone->cow_lock);
 
 	/* Control update. */
+	assert(zone->control_update == NULL || !(zone->control_update->flags & UPDATE_WFEV));
 	zone_control_clear(zone);
 
 	free(zone->catalog_gen);
@@ -237,14 +240,13 @@ void zone_free(zone_t **zone_ptr)
 	pthread_mutex_destroy(&zone->preferred_lock);
 	free(zone->preferred_master);
 
-	/* Free zone contents. */
-	zone_contents_deep_free(zone->contents);
-
 	conf_deactivate_modules(&zone->query_modules, &zone->query_plan);
 
-	ptrlist_free(&zone->reverse_from, NULL);
+	zone_includes_clear(zone);
 	ptrlist_free(&zone->internal_notify, NULL);
 
+	ATOMIC_DEINIT(zone->backup_ctx);
+
 	free(zone);
 	*zone_ptr = NULL;
 }
@@ -461,6 +463,14 @@ bool zone_journal_has_zij(zone_t *zone)
 	return exists && zij;
 }
 
+bool zone_journal_same_serial(zone_t *zone, uint32_t serial_to)
+{
+	bool exists = false;
+	uint32_t journal_serial;
+	int ret = journal_info(zone_journal(zone), &exists, NULL, NULL, &journal_serial, NULL, NULL, NULL, NULL);
+	return ret == KNOT_EOK && exists && journal_serial == serial_to;
+}
+
 void zone_notifailed_clear(zone_t *zone)
 {
 	pthread_mutex_lock(&zone->preferred_lock);
@@ -476,6 +486,16 @@ void zone_schedule_notify(conf_t *conf, zone_t *zone, time_t delay)
 	zone_events_schedule_at(zone, ZONE_EVENT_NOTIFY, time(NULL) + conf_delay + delay);
 }
 
+void zone_schedule_update(conf_t *conf, zone_t *zone, zone_event_type_t type)
+{
+	int64_t conf_delay = 0;
+	if (zone->contents != NULL) {
+		conf_val_t val = conf_zone_get(conf, C_UPDATE_DELAY, zone->name);
+		conf_delay = conf_int(&val);
+	}
+	zone_events_schedule_at(zone, type, time(NULL) + conf_delay);
+}
+
 zone_contents_t *zone_switch_contents(zone_t *zone, zone_contents_t *new_contents)
 {
 	if (zone == NULL) {
@@ -802,7 +822,45 @@ int zone_dump_to_dir(conf_t *conf, zone_t *zone, const char *dir)
 	}
 	free(zonefile);
 
-	return zonefile_write(target, zone->contents);
+	return zonefile_write_skip(target, zone->contents, conf);
+}
+
+bool zone_includes_configured(conf_t *conf, zone_t *zone)
+{
+	return conf_zone_get(conf, C_REVERSE_GEN, zone->name).code == KNOT_EOK ||
+	       conf_zone_get(conf, C_INCLUDE_FROM, zone->name).code == KNOT_EOK;
+}
+
+int zone_includes_add(zone_t *zone, zone_t *include, zone_include_method_t method)
+{
+	zone_include_t *n = calloc(1, sizeof(*n));
+	if (n == NULL) {
+		return KNOT_ENOMEM;
+	}
+	n->include = include;
+	n->method = method;
+	add_tail(&zone->include_from, &n->n);
+	return KNOT_EOK;
+}
+
+void zone_includes_rem(zone_t *zone, zone_t *include)
+{
+	zone_include_t *n, *nxt;
+	WALK_LIST_DELSAFE(n, nxt, zone->include_from) {
+		if (n->include == include) {
+			rem_node(&n->n);
+			free(n);
+		}
+	}
+}
+
+void zone_includes_clear(zone_t *zone)
+{
+	zone_include_t *n, *next;
+	WALK_LIST_DELSAFE(n, next, zone->include_from) {
+		free(n);
+	}
+	init_list(&zone->include_from);
 }
 
 void zone_local_notify_subscribe(zone_t *zone, zone_t *subscribe)
@@ -810,11 +868,17 @@ void zone_local_notify_subscribe(zone_t *zone, zone_t *subscribe)
 	ptrlist_add(&zone->internal_notify, subscribe, NULL);
 }
 
-void zone_local_notify(zone_t *zone)
+void zone_local_notify_unsubscribe(zone_t *zone, zone_t *subscribe)
+{
+	ptrlist_find_rem(&zone->internal_notify, subscribe, NULL);
+}
+
+void zone_local_notify(conf_t *conf, zone_t *zone)
 {
 	ptrnode_t *n;
 	WALK_LIST(n, zone->internal_notify) {
-		zone_events_schedule_now(n->d, ZONE_EVENT_LOAD);
+		zone_t *to_notify = n->d;
+		zone_schedule_update(conf, to_notify, zone_is_slave(conf, to_notify) ? ZONE_EVENT_REFRESH : ZONE_EVENT_LOAD);
 	}
 }
 
@@ -854,7 +918,7 @@ int slave_zone_serial(zone_t *zone, conf_t *conf, uint32_t *serial)
 	*serial = zone_contents_serial(zone->contents);
 
 	conf_val_t val = conf_zone_get(conf, C_DNSSEC_SIGNING, zone->name);
-	if (conf_bool(&val)) {
+	if (conf_bool(&val) || zone_includes_configured(conf, zone)) {
 		ret = zone_get_master_serial(zone, serial);
 	}
 
