@@ -1,34 +1,22 @@
-/*  Copyright (C) 2024 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+/*  Copyright (C) CZ.NIC, z.s.p.o. and contributors
+ *  SPDX-License-Identifier: GPL-2.0-or-later
+ *  For more information, see <https://www.knot-dns.cz/>
  */
 
 #include <assert.h>
-#include <stdio.h>
+#include <inttypes.h>
+#include <libgen.h>
 #include <stdlib.h>
 #include <string.h>
-#include <errno.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <time.h>
 #include <unistd.h>
-#include <inttypes.h>
 
 #include "libknot/libknot.h"
 #include "contrib/files.h"
+#include "contrib/time.h"
 #include "knot/common/log.h"
-#include "knot/dnssec/zone-nsec.h"
 #include "knot/zone/semantic-check.h"
 #include "knot/zone/adjust.h"
 #include "knot/zone/contents.h"
@@ -41,8 +29,8 @@
 
 static void process_error(zs_scanner_t *s)
 {
-	zcreator_t *zc = s->process.data;
-	const knot_dname_t *zname = zc->z->apex->owner;
+	zloader_t *loader = s->process.data;
+	const knot_dname_t *zname = loader->contents->apex->owner;
 
 	ERROR(zname, "%s in zone, file '%s', line %"PRIu64" (%s)",
 	      s->error.fatal ? "fatal error" : "error",
@@ -50,10 +38,8 @@ static void process_error(zs_scanner_t *s)
 	      zs_strerror(s->error.code));
 }
 
-static bool handle_err(zcreator_t *zc, const knot_rrset_t *rr, int ret, bool master)
+static bool handle_err(const knot_dname_t *zname, const knot_rrset_t *rr, int ret)
 {
-	const knot_dname_t *zname = zc->z->apex->owner;
-
 	knot_dname_txt_storage_t buff;
 	char *owner = knot_dname_to_str(buff, rr->owner, sizeof(buff));
 	if (owner == NULL) {
@@ -75,16 +61,19 @@ static bool handle_err(zcreator_t *zc, const knot_rrset_t *rr, int ret, bool mas
 	}
 }
 
-int zcreator_step(zcreator_t *zc, const knot_rrset_t *rr)
+int zcreator_step(zone_contents_t *contents, const knot_rrset_t *rr, zone_skip_t *skip)
 {
-	if (zc == NULL || rr == NULL || rr->rrs.count != 1) {
-		return KNOT_EINVAL;
+	assert(contents);
+	assert(rr);
+
+	if (zone_skip_type(skip, rr->type)) {
+		return KNOT_EOK;
 	}
 
 	zone_node_t *node = NULL;
-	int ret = zone_contents_add_rr(zc->z, rr, &node);
+	int ret = zone_contents_add_rr(contents, rr, &node);
 	if (ret != KNOT_EOK) {
-		if (!handle_err(zc, rr, ret, zc->master)) {
+		if (!handle_err(contents->apex->owner, rr, ret)) {
 			// Fatal error
 			return ret;
 		}
@@ -93,18 +82,18 @@ int zcreator_step(zcreator_t *zc, const knot_rrset_t *rr)
 	return KNOT_EOK;
 }
 
-/*! \brief Creates RR from parser input, passes it to handling function. */
 static void process_data(zs_scanner_t *scanner)
 {
-	zcreator_t *zc = scanner->process.data;
-	if (zc->ret != KNOT_EOK) {
+	zloader_t *zl = scanner->process.data;
+
+	if (zl->ret != KNOT_EOK) {
 		scanner->state = ZS_STATE_STOP;
 		return;
 	}
 
 	knot_dname_t *owner = knot_dname_copy(scanner->r_owner, NULL);
 	if (owner == NULL) {
-		zc->ret = KNOT_ENOMEM;
+		zl->ret = KNOT_ENOMEM;
 		return;
 	}
 
@@ -114,144 +103,171 @@ static void process_data(zs_scanner_t *scanner)
 	int ret = knot_rrset_add_rdata(&rr, scanner->r_data, scanner->r_data_length, NULL);
 	if (ret != KNOT_EOK) {
 		knot_rrset_clear(&rr, NULL);
-		zc->ret = ret;
+		zl->ret = ret;
 		return;
 	}
 
-	/* Convert RDATA dnames to lowercase before adding to zone. */
 	ret = knot_rrset_rr_to_canonical(&rr);
 	if (ret != KNOT_EOK) {
 		knot_rrset_clear(&rr, NULL);
-		zc->ret = ret;
+		zl->ret = ret;
 		return;
 	}
 
-	zc->ret = zcreator_step(zc, &rr);
+	zl->ret = zcreator_step(zl->contents, &rr, zl->skip);
+
 	knot_rrset_clear(&rr, NULL);
 }
 
-int zonefile_open(zloader_t *loader, const char *source, const knot_dname_t *origin,
-                  uint32_t dflt_ttl, semcheck_optional_t semantic_checks, time_t time)
+static void check_origin(zs_scanner_t *s)
 {
-	if (!loader) {
+	if (s->r_type == KNOT_RRTYPE_SOA) {
+		uint8_t *origin_buf = s->process.data;
+		assert(s->r_owner_length <= KNOT_DNAME_MAXLEN);
+		origin_buf[0] = s->r_owner_length;
+		memcpy(origin_buf + 1, s->r_owner, s->r_owner_length);
+		s->state = ZS_STATE_STOP;
+	}
+}
+
+static void error_origin(zs_scanner_t *s)
+{
+	log_error("failed to detect zone origin, file '%s', line %"PRIu64" (%s)",
+	          s->file.name, s->line_counter, zs_strerror(s->error.code));
+}
+
+int zonefile_open(zloader_t *loader, const char *source, const knot_dname_t *origin,
+                  uint32_t dflt_ttl, semcheck_optional_t sem_checks,
+                  sem_handler_t *sem_err_handler, time_t time, zone_skip_t *skip)
+{
+	if (loader == NULL || source == NULL) {
 		return KNOT_EINVAL;
 	}
 
-	memset(loader, 0, sizeof(zloader_t));
-
-	/* Check zone file. */
 	if (access(source, F_OK | R_OK) != 0) {
 		return knot_map_errno();
 	}
 
-	/* Create context. */
-	zcreator_t *zc = malloc(sizeof(zcreator_t));
-	if (zc == NULL) {
-		return KNOT_ENOMEM;
-	}
-	memset(zc, 0, sizeof(zcreator_t));
+	uint8_t origin_buf[1 + KNOT_DNAME_MAXLEN];
+	if (origin == NULL) { // Origin autodetection based on SOA owner and source.
+		const char *ext = ".zone";
+		char *origin_str = basename((char *)source);
+		if (strcmp(origin_str + strlen(origin_str) - strlen(ext), ext) == 0) {
+			origin_str = strndup(origin_str, strlen(origin_str) - strlen(ext));
+		} else {
+			origin_str = strdup(origin_str);
+		}
 
-	zc->z = zone_contents_new(origin, true);
-	if (zc->z == NULL) {
-		free(zc);
-		return KNOT_ENOMEM;
+		origin_buf[0] = 0;
+
+		zs_scanner_t s;
+		if (zs_init(&s, origin_str, KNOT_CLASS_IN, 0) != 0 ||
+		    zs_set_input_file(&s, source) != 0 ||
+		    zs_set_processing(&s, check_origin, error_origin, &origin_buf) != 0) {
+			free(origin_str);
+			zs_deinit(&s);
+			return KNOT_EFILE;
+		}
+		free(origin_str);
+		if (zs_parse_all(&s) != 0 && s.error.fatal) {
+			zs_deinit(&s);
+			return KNOT_EPARSEFAIL;
+		}
+		zs_deinit(&s);
+
+		if (origin_buf[0] == 0) {
+			return KNOT_ESOAINVAL;
+		}
+		origin = origin_buf + 1;
 	}
 
-	/* Prepare textual owner for zone scanner. */
-	char *origin_str = knot_dname_to_str_alloc(origin);
-	if (origin_str == NULL) {
-		zone_contents_deep_free(zc->z);
-		free(zc);
-		return KNOT_ENOMEM;
+	knot_dname_txt_storage_t origin_str;
+	if (knot_dname_to_str(origin_str, origin, sizeof(origin_str)) == NULL) {
+		return KNOT_EINVAL;
 	}
 
 	if (zs_init(&loader->scanner, origin_str, KNOT_CLASS_IN, dflt_ttl) != 0 ||
 	    zs_set_input_file(&loader->scanner, source) != 0 ||
-	    zs_set_processing(&loader->scanner, process_data, process_error, zc) != 0) {
+	    zs_set_processing(&loader->scanner, process_data, process_error, loader) != 0) {
 		zs_deinit(&loader->scanner);
-		free(origin_str);
-		zone_contents_deep_free(zc->z);
-		free(zc);
 		return KNOT_EFILE;
 	}
-	free(origin_str);
+
+	loader->contents = zone_contents_new(origin, true);
+	if (loader->contents == NULL) {
+		zs_deinit(&loader->scanner);
+		return KNOT_ENOMEM;
+	}
 
 	loader->source = strdup(source);
-	loader->creator = zc;
-	loader->semantic_checks = semantic_checks;
+	loader->sem_checks = sem_checks;
+	loader->err_handler = sem_err_handler;
+	loader->skip = skip;
 	loader->time = time;
+	loader->ret = KNOT_EOK;
 
 	return KNOT_EOK;
 }
 
 zone_contents_t *zonefile_load(zloader_t *loader, uint16_t threads)
 {
-	if (!loader) {
+	if (loader == NULL) {
 		return NULL;
 	}
 
-	zcreator_t *zc = loader->creator;
-	const knot_dname_t *zname = zc->z->apex->owner;
+	const knot_dname_t *zname = loader->contents->apex->owner;
 
-	assert(zc);
 	int ret = zs_parse_all(&loader->scanner);
 	if (ret != 0 && loader->scanner.error.counter == 0) {
-		ERROR(zname, "failed to load zone, file '%s' (%s)",
+		ERROR(zname, "failed to load, file '%s' (%s)",
 		      loader->source, zs_strerror(loader->scanner.error.code));
 		goto fail;
-	}
-
-	if (zc->ret != KNOT_EOK) {
-		ERROR(zname, "failed to load zone, file '%s' (%s)",
-		      loader->source, knot_strerror(zc->ret));
+	} else if (loader->ret != KNOT_EOK) {
+		ERROR(zname, "failed to load, file '%s' (%s)",
+		      loader->source, knot_strerror(loader->ret));
 		goto fail;
-	}
-
-	if (loader->scanner.error.counter > 0) {
-		ERROR(zname, "failed to load zone, file '%s', %"PRIu64" errors",
+	} else if (loader->scanner.error.counter > 0) {
+		ERROR(zname, "failed to load, file '%s', %"PRIu64" errors",
 		      loader->source, loader->scanner.error.counter);
 		goto fail;
 	}
 
-	knot_rdataset_t *soa = node_rdataset(zc->z->apex, KNOT_RRTYPE_SOA);
+	knot_rdataset_t *soa = node_rdataset(loader->contents->apex, KNOT_RRTYPE_SOA);
 	if (soa == NULL || soa->count != 1) {
 		sem_error_t code = (soa == NULL) ? SEM_ERR_SOA_NONE : SEM_ERR_SOA_MULTIPLE;
 		loader->err_handler->error = true;
-		loader->err_handler->cb(loader->err_handler, zc->z, NULL, code, NULL);
+		loader->err_handler->cb(loader->err_handler, loader->contents, NULL, code, NULL);
 		goto fail;
 	}
 
-	ret = zone_adjust_contents(zc->z, adjust_cb_flags_and_nsec3, adjust_cb_nsec3_flags,
-	                           true, true, 1, NULL);
+	ret = zone_adjust_contents(loader->contents, adjust_cb_flags_and_nsec3,
+	                           adjust_cb_nsec3_flags, true, true, 1, NULL);
 	if (ret != KNOT_EOK) {
-		ERROR(zname, "failed to finalize zone contents (%s)",
-		      knot_strerror(ret));
+		ERROR(zname, "failed to finalize zone contents (%s)", knot_strerror(ret));
 		goto fail;
 	}
 
-	ret = sem_checks_process(zc->z, loader->semantic_checks,
+	ret = sem_checks_process(loader->contents, loader->sem_checks,
 	                         loader->err_handler, loader->time, threads);
 
 	if (ret != KNOT_EOK) {
-		ERROR(zname, "failed to load zone, file '%s' (%s)",
-		      loader->source, knot_strerror(ret));
+		ERROR(zname, "failed to check zone (%s)", knot_strerror(ret));
 		goto fail;
 	}
 
 	/* The contents will now change possibly messing up NSEC3 tree, it will
 	   be adjusted again at zone_update_commit. */
-	ret = zone_adjust_contents(zc->z, unadjust_cb_point_to_nsec3, NULL,
-	                           false, false, 1, NULL);
+	ret = zone_adjust_contents(loader->contents, unadjust_cb_point_to_nsec3,
+	                           NULL, false, false, 1, NULL);
 	if (ret != KNOT_EOK) {
-		ERROR(zname, "failed to finalize zone contents (%s)",
-		      knot_strerror(ret));
+		ERROR(zname, "failed to finalize zone contents (%s)", knot_strerror(ret));
 		goto fail;
 	}
 
-	return zc->z;
-
+	return loader->contents;
 fail:
-	zone_contents_deep_free(zc->z);
+	zone_contents_deep_free(loader->contents);
+
 	return NULL;
 }
 
@@ -273,7 +289,7 @@ int zonefile_exists(const char *path, struct timespec *mtime)
 	return KNOT_EOK;
 }
 
-int zonefile_write(const char *path, zone_contents_t *zone)
+int zonefile_write(const char *path, zone_contents_t *zone, zone_skip_t *skip)
 {
 	if (path == NULL) {
 		return KNOT_EINVAL;
@@ -297,7 +313,7 @@ int zonefile_write(const char *path, zone_contents_t *zone)
 		return ret;
 	}
 
-	ret = zone_dump_text(zone, file, true, NULL);
+	ret = zone_dump_text(zone, skip, file, true, NULL);
 	fclose(file);
 	if (ret != KNOT_EOK) {
 		unlink(tmp_name);
@@ -321,13 +337,12 @@ int zonefile_write(const char *path, zone_contents_t *zone)
 
 void zonefile_close(zloader_t *loader)
 {
-	if (!loader) {
+	if (loader == NULL) {
 		return;
 	}
 
 	zs_deinit(&loader->scanner);
 	free(loader->source);
-	free(loader->creator);
 }
 
 void err_handler_logger(sem_handler_t *handler, const zone_contents_t *zone,
