@@ -1,17 +1,6 @@
-/*  Copyright (C) 2020 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+/*  Copyright (C) CZ.NIC, z.s.p.o. and contributors
+ *  SPDX-License-Identifier: GPL-2.0-or-later
+ *  For more information, see <https://www.knot-dns.cz/>
  */
 
 /*!
@@ -48,7 +37,7 @@ typedef char knot_dname_txt_storage_t[KNOT_DNAME_TXT_MAXLEN + 1];
  *
  * \param name  Name on the wire.
  * \param endp  Name boundary.
- * \param pkt   Wire.
+ * \param pkt   Wire (can be NULL if not compressed).
  *
  * \retval (compressed) size of the domain name.
  * \retval KNOT_EINVAL
@@ -161,12 +150,22 @@ static inline knot_dname_t *knot_dname_from_str_alloc(const char *name)
 }
 
 /*!
- * \brief Convert name to lowercase.
+ * \brief Convert domain name to lowercase.
  *
  * \param name  Domain name to be converted.
  */
 void knot_dname_to_lower(knot_dname_t *name);
 
+/*!
+ * \brief Copy lowercased domain name.
+ *
+ * \note The output buffer isn't checked if it's big enough!
+ *
+ * \param dst   Destination buffer.
+ * \param name  Source domain name to be converted.
+ */
+void knot_dname_copy_lower(knot_dname_t *dst, const knot_dname_t *name);
+
 /*!
  * \brief Returns size of the given domain name.
  *
@@ -185,7 +184,7 @@ size_t knot_dname_size(const knot_dname_t *name);
  * \brief Returns full size of the given domain name (expanded compression ptrs).
  *
  * \param name  Domain name to get the size of.
- * \param pkt   Related packet (or NULL if unpacked)
+ * \param pkt   Related packet.
  *
  * \retval size of the domain name.
  * \retval 0 if invalid argument.
@@ -280,17 +279,27 @@ bool knot_dname_is_equal(const knot_dname_t *d1, const knot_dname_t *d2);
 _pure_
 bool knot_dname_is_case_equal(const knot_dname_t *d1, const knot_dname_t *d2);
 
+/*!
+ * \brief Checks the domain name if it contains '\0' byte within a label.
+ *
+ * \param name  Domain name.
+ *
+ * \retval true if a zero byte is included
+ * \retval false if a zero byte isn't included
+ */
+_pure_
+bool knot_dname_with_null(const knot_dname_t *name);
+
 /*!
  * \brief Count length of the N first labels.
  *
  * \param name     Domain name.
  * \param nlabels  First N labels.
- * \param pkt      Related packet (or NULL if not compressed).
  *
  * \return Length of the prefix.
  */
 _pure_
-size_t knot_dname_prefixlen(const uint8_t *name, unsigned nlabels, const uint8_t *pkt);
+size_t knot_dname_prefixlen(const uint8_t *name, unsigned nlabels);
 
 /*!
  * \brief Return number of labels in the domain name.
