@@ -1,17 +1,6 @@
-/*  Copyright (C) 2022 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+/*  Copyright (C) CZ.NIC, z.s.p.o. and contributors
+ *  SPDX-License-Identifier: GPL-2.0-or-later
+ *  For more information, see <https://www.knot-dns.cz/>
  */
 
 /*!
@@ -40,6 +29,7 @@ typedef enum {
 	KNOT_XDP_MSG_RST   = (1 << 5), /*!< RST flag set (TCP only). */
 	KNOT_XDP_MSG_MSS   = (1 << 6), /*!< MSS option in TCP header (TCP only). */
 	KNOT_XDP_MSG_WSC   = (1 << 7), /*!< Window Scale option in TCP header. */
+	KNOT_XDP_MSG_VLAN  = (1 << 8), /*!< This packet will contain VLAN header. */
 } knot_xdp_msg_flag_t;
 
 /*! \brief Packet description with src & dst MAC & IP addrs + DNS payload. */
@@ -55,6 +45,8 @@ typedef struct knot_xdp_msg {
 	uint16_t mss;
 	uint16_t win;
 	uint8_t win_scale;
+	uint8_t ecn;
+	uint16_t vlan_tci;
 } knot_xdp_msg_t;
 
 /*! @} */
