@@ -1,17 +1,6 @@
-/*  Copyright (C) 2022 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+/*  Copyright (C) CZ.NIC, z.s.p.o. and contributors
+ *  SPDX-License-Identifier: GPL-2.0-or-later
+ *  For more information, see <https://www.knot-dns.cz/>
  */
 
 #include <stdint.h>
@@ -819,9 +808,9 @@ int date_to_timestamp(uint8_t *buff, uint32_t *timestamp)
 	return ZS_OK;
 }
 
-void wire_dname_to_str(const uint8_t  *data,
-                       const uint32_t data_len,
-                       char           *text)
+void dname_to_str(const uint8_t  *data,
+                  const uint32_t data_len,
+                  char           *text)
 {
 	uint32_t i = 0, text_len = 0;
 
