@@ -1,17 +1,6 @@
-/*  Copyright (C) 2021 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+/*  Copyright (C) CZ.NIC, z.s.p.o. and contributors
+ *  SPDX-License-Identifier: GPL-2.0-or-later
+ *  For more information, see <https://www.knot-dns.cz/>
  */
 
 /*!
@@ -157,6 +146,8 @@ struct zs_scanner {
 	bool     long_string;
 	/*! Comma separated string list indication (svcb parsing). */
 	bool     comma_list;
+	/*! Indication of a non-applied backslash. */
+	bool     pending_backslash;
 
 	/*! Pointer to the actual dname storage (origin/owner/rdata). */
 	uint8_t  *dname;
@@ -274,7 +265,7 @@ struct zs_scanner {
  * \note Error code is stored in the scanner context.
  *
  * \param scanner  Scanner context.
- * \param origin   Initial zone origin.
+ * \param origin   Initial zone origin ("." is used if empty).
  * \param rclass   Zone class value.
  * \param ttl      Initial ttl value.
  *
