@@ -1,17 +1,6 @@
-/*  Copyright (C) 2024 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+/*  Copyright (C) CZ.NIC, z.s.p.o. and contributors
+ *  SPDX-License-Identifier: GPL-2.0-or-later
+ *  For more information, see <https://www.knot-dns.cz/>
  */
 
 #include <getopt.h>
@@ -74,6 +63,9 @@ static void print_help(void)
 	       "                 (syntax: import-pkcs11 <key_id> <attribute_name>=<value>...)\n"
 	       "  nsec3-salt    Print current NSEC3 salt. If a parameter is specified, set new salt.\n"
 	       "                 (syntax: nsec3-salt [<new_salt>])\n"
+	       "  local-serial  Print SOA serial stored in KASP database when using on-slave signing.\n"
+	       "                 If a parameter is specified, set new serial.\n"
+	       "                 (syntax: local-serial <new_serial>)\n"
 	       "  master-serial Print SOA serial of the remote master stored in KASP database when using on-slave signing.\n"
 	       "                 If a parameter is specified, set new master serial.\n"
 	       "                 (syntax: master-serial <new_serial>)\n"
@@ -192,8 +184,8 @@ static int key_command(int argc, char *argv[], int opt_ind, knot_lmdb_db_t *kasp
 			ret = keymgr_nsec3_salt_print(&kctx);
 			print_ok_on_succes = false;
 		}
-	} else if (same_command(argv[1], "master-serial", false)) {
-		kaspdb_serial_t type = KASPDB_SERIAL_MASTER;
+	} else if (same_command(argv[1], "local-serial", false) || same_command(argv[1], "master-serial", false)) {
+		kaspdb_serial_t type = (argv[1][0] == 'm' ? KASPDB_SERIAL_MASTER : KASPDB_SERIAL_LASTSIGNED);
 		if (argc > 2) {
 			uint32_t new_serial = 0;
 			if ((ret = str_to_u32(argv[2], &new_serial)) == KNOT_EOK) {
@@ -203,9 +195,6 @@ static int key_command(int argc, char *argv[], int opt_ind, knot_lmdb_db_t *kasp
 			ret = keymgr_serial_print(&kctx, type);
 			print_ok_on_succes = false;
 		}
-	} else if (same_command(argv[1], "local-serial", false)) {
-		WARN2("deprecated command with no effect");
-		print_ok_on_succes = false;
 	} else if (same_command(argv[1], "set", false)) {
 		CHECK_MISSING_ARG("Key is not specified");
 		knot_kasp_key_t *key2set;
