@@ -1,17 +1,6 @@
-/*  Copyright (C) 2024 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+/*  Copyright (C) CZ.NIC, z.s.p.o. and contributors
+ *  SPDX-License-Identifier: GPL-2.0-or-later
+ *  For more information, see <https://www.knot-dns.cz/>
  */
 
 #include <dirent.h>
@@ -35,7 +24,7 @@
 #include "contrib/strtonum.h"
 #include "contrib/threads.h"
 #include "contrib/time.h"
-#include "knot/ctl/process.h"
+#include "knot/ctl/threads.h"
 #include "knot/conf/conf.h"
 #include "knot/conf/migration.h"
 #include "knot/conf/module.h"
@@ -45,36 +34,16 @@
 #include "knot/common/stats.h"
 #include "knot/common/systemd.h"
 #include "knot/server/server.h"
+#include "knot/server/signals.h"
 #include "knot/server/tcp-handler.h"
 #include "utils/common/params.h"
 
 #define PROGRAM_NAME "knotd"
 
-typedef enum {
-	CONCURRENT_EMPTY = 0,   // fresh cctx without a thread.
-	CONCURRENT_ASSIGNED,    // cctx assigned to process a command.
-	CONCURRENT_RUNNING,     // ctl command is being processed in the thread.
-	CONCURRENT_IDLE,        // command has been processed, waiting for a new one.
-	CONCURRENT_KILLED,      // cctx cleanup has started.
-	CONCURRENT_FINISHED,    // after having been killed, the thread is being joined.
-} concurrent_ctl_state_t;
-
 typedef struct {
-	concurrent_ctl_state_t state;
-	pthread_mutex_t mutex;  // Protects .state.
-	pthread_cond_t cond;
-	knot_ctl_t *ctl;
-	server_t *server;
-	pthread_t thread;
-	int ret;
-	int thread_idx;
-	bool exclusive;
-} concurrent_ctl_ctx_t;
-
-/* Signal flags. */
-static volatile bool sig_req_stop = false;
-static volatile bool sig_req_reload = false;
-static volatile bool sig_req_zones_reload = false;
+	bool            wait_set;
+	knot_millis_t   loop_wait;
+} startup_data_t;
 
 static int make_daemon(int nochdir, int noclose)
 {
@@ -133,83 +102,6 @@ static int make_daemon(int nochdir, int noclose)
 	return 0;
 }
 
-struct signal {
-	int signum;
-	bool handle;
-};
-
-/*! \brief Signals used by the server. */
-static const struct signal SIGNALS[] = {
-	{ SIGHUP,  true  },  /* Reload server. */
-	{ SIGUSR1, true  },  /* Reload zones. */
-	{ SIGINT,  true  },  /* Terminate server. */
-	{ SIGTERM, true  },  /* Terminate server. */
-	{ SIGALRM, false },  /* Internal thread synchronization. */
-	{ SIGPIPE, false },  /* Ignored. Some I/O errors. */
-	{ 0 }
-};
-
-/*! \brief Server signal handler. */
-static void handle_signal(int signum)
-{
-	switch (signum) {
-	case SIGHUP:
-		sig_req_reload = true;
-		break;
-	case SIGUSR1:
-		sig_req_zones_reload = true;
-		break;
-	case SIGINT:
-	case SIGTERM:
-		if (sig_req_stop) {
-			exit(EXIT_FAILURE);
-		}
-		sig_req_stop = true;
-		break;
-	default:
-		/* ignore */
-		break;
-	}
-}
-
-/*! \brief Setup signal handlers and blocking mask. */
-static void setup_signals(void)
-{
-	/* Block all signals. */
-	static sigset_t all;
-	sigfillset(&all);
-	sigdelset(&all, SIGPROF);
-	sigdelset(&all, SIGQUIT);
-	sigdelset(&all, SIGILL);
-	sigdelset(&all, SIGABRT);
-	sigdelset(&all, SIGBUS);
-	sigdelset(&all, SIGFPE);
-	sigdelset(&all, SIGSEGV);
-
-	/* Setup handlers. */
-	struct sigaction action = { .sa_handler = handle_signal };
-	for (const struct signal *s = SIGNALS; s->signum > 0; s++) {
-		sigaction(s->signum, &action, NULL);
-	}
-
-	pthread_sigmask(SIG_SETMASK, &all, NULL);
-}
-
-/*! \brief Unblock server control signals. */
-static void enable_signals(void)
-{
-	sigset_t mask;
-	sigemptyset(&mask);
-
-	for (const struct signal *s = SIGNALS; s->signum > 0; s++) {
-		if (s->handle) {
-			sigaddset(&mask, s->signum);
-		}
-	}
-
-	pthread_sigmask(SIG_UNBLOCK, &mask, NULL);
-}
-
 /*! \brief Drop POSIX 1003.1e capabilities. */
 static void drop_capabilities(void)
 {
@@ -229,309 +121,272 @@ static void drop_capabilities(void)
 #endif /* ENABLE_CAP_NG */
 }
 
-static void check_loaded(server_t *server)
+/* The following time constants are in milliseconds. */
+#define LOOP_WAIT           5000LLU  /* After loading all zones. It could be even infinity. */
+#define LOOP_WAIT_MIN        200LLU  /* Minimum wait before all zones are started. */
+#define LOOP_WAIT_MAX      30000LLU  /* Maximum wait before all zones are started. */
+#define LOOP_WAIT_LOAD_MIN  1000LLU  /* Minimum wait before all zones are loaded, after 'started'. */
+#define LOOP_WAIT_LOAD_MAX 10000LLU  /* Can be overruled by higher initial wait value. */
+#define LOAD_START_DECLINE 60000LLU  /* Before the wait starts to grow. */
+
+/*! \brief Calculate wait in 'started' state detection. */
+static void calc_wait(startup_data_t *startup_data, knot_millis_t t)
 {
-	static bool finished = false;
-	if (finished) {
+	/* Multiply wait by 2 -- just a chosen number. */
+	startup_data->loop_wait = MAX(LOOP_WAIT_MIN, MIN(LOOP_WAIT_MAX, t * 2));
+}
+
+/*! \brief Calculate wait in 'loaded' state detection. */
+static void calc_wait_loaded(startup_data_t *startup_data, knot_millis_t t)
+{
+	if (t < LOAD_START_DECLINE) {
+		startup_data->loop_wait = MAX(LOOP_WAIT_LOAD_MIN, startup_data->loop_wait);
 		return;
 	}
 
-	/* Avoid traversing the zonedb too frequently. */
-	static struct timespec last = { 0 };
-	struct timespec now = time_now();
-	if (last.tv_sec == now.tv_sec) {
-		return;
-	}
-	last = now;
+	/* Multiply wait by approx. 1.05 -- just a chosen number. */
+	startup_data->loop_wait = MIN(LOOP_WAIT_LOAD_MAX,
+	                              startup_data->loop_wait + startup_data->loop_wait / 20);
+}
 
-	if (!(conf()->cache.srv_dbus_event & DBUS_EVENT_RUNNING)) {
-		finished = true;
-		return;
-	}
+static void walk_zonedb(server_t *server, bool *out_start, bool *out_load,
+                        bool started, bool fast)
+{
+	bool start = true;
+	bool load = true;
 
 	rcu_read_lock();
 	knot_zonedb_iter_t *it = knot_zonedb_iter_begin(server->zone_db);
 	while (!knot_zonedb_iter_finished(it)) {
 		zone_t *zone = (zone_t *)knot_zonedb_iter_val(it);
 		if (zone->contents == NULL) {
-			knot_zonedb_iter_free(it);
-			rcu_read_unlock();
-			return;
+			load = false;
+			if (started && fast) {
+				break;
+			} else if (!zone->started) {
+				start = false;
+				if (fast) {
+					break;
+				}
+			}
 		}
 		knot_zonedb_iter_next(it);
 	}
 	knot_zonedb_iter_free(it);
 	rcu_read_unlock();
 
-	finished = true;
-	dbus_emit_running(true);
+	*out_start = start;
+	*out_load = load;
 }
 
-static void *ctl_process_thread(void *arg);
-
-/*!
- * Try to find an empty ctl processing context and if successful,
- * prepare to lauch the incomming command processing in it.
- *
- * \param[in]  concurrent_ctxs  Configured concurrent control contexts.
- * \param[in]  n_ctxs           Number of configured concurrent control contexts.
- * \param[in]  ctl              Control context.
- *
- * \return     Assigned concurrent control context, or NULL.
- */
-
-static concurrent_ctl_ctx_t *find_free_ctx(concurrent_ctl_ctx_t *concurrent_ctxs,
-                                           size_t n_ctxs, knot_ctl_t *ctl)
+static int check_loaded(server_t *server, bool async, startup_data_t *startup_data)
 {
-	concurrent_ctl_ctx_t *res = NULL;
-	for (size_t i = 0; i < n_ctxs && res == NULL; i++) {
-		concurrent_ctl_ctx_t *cctx = &concurrent_ctxs[i];
-		pthread_mutex_lock(&cctx->mutex);
-		if (cctx->exclusive) {
-			while (cctx->state != CONCURRENT_IDLE) {
-				pthread_cond_wait(&cctx->cond, &cctx->mutex);
-			}
-			knot_ctl_free(cctx->ctl);
-			cctx->ctl = knot_ctl_clone(ctl);
-			if (cctx->ctl == NULL) {
-				cctx->exclusive = false;
-				pthread_mutex_unlock(&cctx->mutex);
-				break;
-			}
-			cctx->state = CONCURRENT_ASSIGNED;
-			res = cctx;
-			pthread_cond_broadcast(&cctx->cond);
-		}
-		pthread_mutex_unlock(&cctx->mutex);
+	/*
+	 * Started: all zones loaded or at least tried to load at least once,
+	 *          the server is already running and accepting queries.
+	 * Loaded:  all zones successfully loaded, it implies 'started',
+	 *          KNOT_BUS_EVENT_STARTED already emitted over DBus.
+	 */
+	static bool started = false;
+	static bool loaded = false;
+	static knot_millis_t started_time = 0;
+	assert(server->state & ServerRunning);
+	if (loaded) {
+		assert(server->state & ServerAnswering);
+		return KNOT_EOK;
 	}
-	for (size_t i = 0; i < n_ctxs && res == NULL; i++) {
-		concurrent_ctl_ctx_t *cctx = &concurrent_ctxs[i];
-		pthread_mutex_lock(&cctx->mutex);
-		switch (cctx->state) {
-		case CONCURRENT_EMPTY:
-			(void)thread_create_nosignal(&cctx->thread, ctl_process_thread, cctx);
-			break;
-		case CONCURRENT_IDLE:
-			knot_ctl_free(cctx->ctl);
-			pthread_cond_broadcast(&cctx->cond);
-			break;
-		default:
-			pthread_mutex_unlock(&cctx->mutex);
-			continue;
+
+	bool start, load;
+	knot_millis_t now_ms = knot_millis_now();
+
+	/* In the first db walk, benchmark the zonedb traversal while checking. */
+	walk_zonedb(server, &start, &load, started, startup_data->wait_set);
+
+	if (!startup_data->wait_set) {
+		knot_millis_t afterwalk = knot_millis_now();
+		calc_wait(startup_data, afterwalk - now_ms);
+		if (async) {
+			calc_wait_loaded(startup_data, 0);
 		}
-		cctx->ctl = knot_ctl_clone(ctl);
-		if (cctx->ctl != NULL) {
-			cctx->state = CONCURRENT_ASSIGNED;
-			res = cctx;
-		}
-		pthread_mutex_unlock(&cctx->mutex);
+		startup_data->wait_set = true;
 	}
+
+	if (!start) {
+		return KNOT_EOK;
+	}
+	if (!started) {
+		started_time = now_ms;
+		if (!async) {
+			int ret = server_start_answering(server);
+			if (ret != KNOT_EOK) {
+				return ret;
+			}
+		}
+		started = true;
+	}
+
+	assert(server->state & ServerAnswering);
+	if (!(conf()->cache.srv_dbus_event & DBUS_EVENT_RUNNING)) {
+		loaded = true;
+		return KNOT_EOK;
+	}
+
+	if (load) {   /* Not 'loaded' yet. */
+		dbus_emit_running(true);
+		loaded = true;
+		startup_data->loop_wait = LOOP_WAIT;
+	} else {
+		/* Extend the wait gradually. */
+		calc_wait_loaded(startup_data, now_ms - started_time);
+	}
+
+	return KNOT_EOK;
+}
+
+static void deinit_ctls(knot_ctl_t **ctls, unsigned count)
+{
+	for (unsigned i = 0; i < count; i++) {
+		knot_ctl_unbind(ctls[i]);
+		knot_ctl_free(ctls[i]);
+	}
+	free(ctls);
+}
+
+static unsigned count_ctls(const char *socket, conf_val_t *listen_val)
+{
+	return (socket == NULL) ? MAX(1, conf_val_count(listen_val)) : 1;
+}
+
+static knot_ctl_t **init_ctls(const char *socket)
+{
+	conf_val_t listen_val = conf_get(conf(), C_CTL, C_LISTEN);
+	unsigned cnt = count_ctls(socket, &listen_val);
+
+	knot_ctl_t **res = calloc(cnt, sizeof(*res));
+	for (unsigned i = 0; i < cnt; i++) {
+		res[i] = knot_ctl_alloc();
+		if (res[i] == NULL) {
+			log_fatal("control, failed to initialize socket");
+			deinit_ctls(res, i);
+			return NULL;
+		}
+	}
+
+	uint16_t backlog = conf_get_int(conf(), C_CTL, C_BACKLOG);
+	conf_val_t rundir_val = conf_get(conf(), C_SRV, C_RUNDIR);
+	char *rundir = conf_abs_path(&rundir_val, NULL);
+
+	int ret = KNOT_EOK;
+	for (unsigned i = 0; i < cnt && ret == KNOT_EOK; i++) {
+		char *listen = (socket == NULL) ? conf_abs_path(&listen_val, rundir)
+		                                : strdup(socket);
+		if (listen == NULL) {
+			log_fatal("control, empty socket path");
+			ret = KNOT_ENOENT;
+		} else {
+			knot_ctl_set_timeout(res[i], conf()->cache.ctl_timeout);
+			log_info("control, binding to '%s'", listen);
+			ret = knot_ctl_bind(res[i], listen, backlog);
+			if (ret != KNOT_EOK) {
+				log_fatal("control, failed to bind socket '%s' (%s)",
+				          listen, knot_strerror(ret));
+			}
+			free(listen);
+		}
+		if (cnt > 1) {
+			conf_val_next(&listen_val);
+		}
+	}
+	if (ret != KNOT_EOK) {
+		deinit_ctls(res, cnt);
+		res = NULL;
+	}
+	free(rundir);
+
 	return res;
 }
 
-static void init_ctxs(concurrent_ctl_ctx_t *concurrent_ctxs, size_t n_ctxs, server_t *server)
-{
-	for (size_t i = 0; i < n_ctxs; i++) {
-		concurrent_ctl_ctx_t *cctx = &concurrent_ctxs[i];
-		pthread_mutex_init(&cctx->mutex, NULL);
-		pthread_cond_init(&cctx->cond, NULL);
-		cctx->server = server;
-		cctx->thread_idx = i + 1;
-	}
-}
-
-static int cleanup_ctxs(concurrent_ctl_ctx_t *concurrent_ctxs, size_t n_ctxs)
-{
-	int ret = KNOT_EOK;
-	for (size_t i = 0; i < n_ctxs; i++) {
-		concurrent_ctl_ctx_t *cctx = &concurrent_ctxs[i];
-		pthread_mutex_lock(&cctx->mutex);
-		if (cctx->state == CONCURRENT_IDLE) {
-			knot_ctl_free(cctx->ctl);
-			cctx->ctl = NULL;
-			if (cctx->ret == KNOT_CTL_ESTOP) {
-				ret = cctx->ret;
-			}
-		}
-		pthread_mutex_unlock(&cctx->mutex);
-	}
-	return ret;
-}
-
-static void finalize_ctxs(concurrent_ctl_ctx_t *concurrent_ctxs, size_t n_ctxs)
-{
-	for (size_t i = 0; i < n_ctxs; i++) {
-		concurrent_ctl_ctx_t *cctx = &concurrent_ctxs[i];
-		pthread_mutex_lock(&cctx->mutex);
-		if (cctx->state == CONCURRENT_EMPTY) {
-			pthread_mutex_unlock(&cctx->mutex);
-			pthread_mutex_destroy(&cctx->mutex);
-			pthread_cond_destroy(&cctx->cond);
-			continue;
-		}
-
-		cctx->state = CONCURRENT_KILLED;
-		pthread_cond_broadcast(&cctx->cond);
-		pthread_mutex_unlock(&cctx->mutex);
-		(void)pthread_join(cctx->thread, NULL);
-
-		assert(cctx->state == CONCURRENT_FINISHED);
-		knot_ctl_free(cctx->ctl);
-		pthread_mutex_destroy(&cctx->mutex);
-		pthread_cond_destroy(&cctx->cond);
-	}
-}
-
-static void *ctl_process_thread(void *arg)
-{
-	concurrent_ctl_ctx_t *ctx = arg;
-	rcu_register_thread();
-	setup_signals(); // in fact, this blocks common signals so that they
-	                 // arrive to main thread instead of this one
-
-	pthread_mutex_lock(&ctx->mutex);
-	while (ctx->state != CONCURRENT_KILLED) {
-		if (ctx->state != CONCURRENT_ASSIGNED) {
-			pthread_cond_wait(&ctx->cond, &ctx->mutex);
-			continue;
-		}
-		ctx->state = CONCURRENT_RUNNING;
-		bool exclusive = ctx->exclusive;
-		pthread_mutex_unlock(&ctx->mutex);
-
-		// Not IDLE, ctx can be read without locking.
-		int ret = ctl_process(ctx->ctl, ctx->server, ctx->thread_idx, &exclusive);
-
-		pthread_mutex_lock(&ctx->mutex);
-		ctx->ret = ret;
-		ctx->exclusive = exclusive;
-		if (ctx->state == CONCURRENT_RUNNING) { // not KILLED
-			ctx->state = CONCURRENT_IDLE;
-			pthread_cond_broadcast(&ctx->cond);
-		}
-	}
-
-	knot_ctl_close(ctx->ctl);
-
-	ctx->state = CONCURRENT_FINISHED;
-	pthread_mutex_unlock(&ctx->mutex);
-	rcu_unregister_thread();
-	return NULL;
-}
-
 /*! \brief Event loop listening for signals and remote commands. */
-static void event_loop(server_t *server, const char *socket, bool daemonize,
-                       unsigned long pid)
+static int event_loop(server_t *server, const char *socket, bool daemonize,
+                      unsigned long pid, bool async)
 {
-	knot_ctl_t *ctl = knot_ctl_alloc();
-	if (ctl == NULL) {
-		log_fatal("control, failed to initialize (%s)",
-		          knot_strerror(KNOT_ENOMEM));
-		return;
+	knot_ctl_t **ctls = init_ctls(socket);
+	if (ctls == NULL) {
+		return KNOT_ERROR;
 	}
 
-	// Set control timeout.
-	knot_ctl_set_timeout(ctl, conf()->cache.ctl_timeout);
+	conf_val_t listen_val = conf_get(conf(), C_CTL, C_LISTEN);
+	unsigned sock_count = count_ctls(socket, &listen_val);
+	ctl_socket_ctx_t sctx = {
+		.ctls = ctls,
+		.server = server,
+		.thr_count = CTL_MAX_CONCURRENT / sock_count
+	};
 
-	/* Get control socket configuration. */
-	char *listen;
-	if (socket == NULL) {
-		conf_val_t listen_val = conf_get(conf(), C_CTL, C_LISTEN);
-		conf_val_t rundir_val = conf_get(conf(), C_SRV, C_RUNDIR);
-		char *rundir = conf_abs_path(&rundir_val, NULL);
-		listen = conf_abs_path(&listen_val, rundir);
-		free(rundir);
-	} else {
-		listen = strdup(socket);
-	}
-	if (listen == NULL) {
-		knot_ctl_free(ctl);
-		log_fatal("control, empty socket path");
-		return;
-	}
-
-	log_info("control, binding to '%s'", listen);
-
-	/* Bind the control socket. */
-	uint16_t backlog = conf_get_int(conf(), C_CTL, C_BACKLOG);
-	int ret = knot_ctl_bind(ctl, listen, backlog);
+	int ret = ctl_socket_thr_init(&sctx, sock_count);
 	if (ret != KNOT_EOK) {
-		knot_ctl_free(ctl);
-		log_fatal("control, failed to bind socket '%s' (%s)",
-		          listen, knot_strerror(ret));
-		free(listen);
-		return;
+		log_fatal("control, failed to launch socket threads (%s)",
+		          knot_strerror(ret));
+		return ret;
 	}
-	free(listen);
 
-	enable_signals();
-
-	concurrent_ctl_ctx_t concurrent_ctxs[CTL_MAX_CONCURRENT] = { 0 };
-	init_ctxs(concurrent_ctxs, CTL_MAX_CONCURRENT, server);
-	bool main_thread_exclusive = false;
+	signals_enable();
 
 	/* Notify systemd about successful start. */
 	systemd_ready_notify();
 	if (daemonize) {
-		log_info("server started as a daemon, PID %lu", pid);
+		log_info("starting server as a daemon, PID %lu", pid);
 	} else {
-		log_info("server started in the foreground, PID %lu", pid);
+		log_info("starting server in the foreground, PID %lu", pid);
 	}
 
-	/* Run event loop. */
+	/* Server startup parameters. */
+	startup_data_t startup_data = {
+		.wait_set = false
+	};
+
+	/* Run interrupt processing loop. */
 	for (;;) {
-		/* Interrupts. */
-		if (sig_req_reload && !sig_req_stop) {
-			sig_req_reload = false;
+		if (signals_req_reload && !signals_req_stop) {
+			signals_req_reload = false;
 			pthread_rwlock_wrlock(&server->ctl_lock);
 			server_reload(server, RELOAD_FULL);
 			pthread_rwlock_unlock(&server->ctl_lock);
 		}
-		if (sig_req_zones_reload && !sig_req_stop) {
-			sig_req_zones_reload = false;
-			reload_t mode = server->catalog_upd_signal ? RELOAD_CATALOG : RELOAD_ZONES;
+		if (signals_req_zones_reload && !signals_req_stop) {
+			signals_req_zones_reload = false;
+			reload_t mode = ATOMIC_GET(server->catalog_upd_signal) ?
+			                RELOAD_CATALOG : RELOAD_FULL;
 			pthread_rwlock_wrlock(&server->ctl_lock);
-			server->catalog_upd_signal = false;
+			ATOMIC_SET(server->catalog_upd_signal, false);
 			server_update_zones(conf(), server, mode);
 			pthread_rwlock_unlock(&server->ctl_lock);
 		}
-		if (sig_req_stop || cleanup_ctxs(concurrent_ctxs, CTL_MAX_CONCURRENT) == KNOT_CTL_ESTOP) {
+		if (signals_req_stop) {
 			break;
 		}
 
-		// Update control timeout.
-		knot_ctl_set_timeout(ctl, conf()->cache.ctl_timeout);
-
-		if (sig_req_reload || sig_req_zones_reload) {
+		if (signals_req_reload || signals_req_zones_reload) {
 			continue;
 		}
 
-		check_loaded(server);
-
-		ret = knot_ctl_accept(ctl);
+		ret = check_loaded(server, async, &startup_data);
 		if (ret != KNOT_EOK) {
-			continue;
+			goto done;
 		}
 
-		if (main_thread_exclusive ||
-		    find_free_ctx(concurrent_ctxs, CTL_MAX_CONCURRENT, ctl) == NULL) {
-			ret = ctl_process(ctl, server, 0, &main_thread_exclusive);
-			knot_ctl_close(ctl);
-			if (ret == KNOT_CTL_ESTOP) {
-				break;
-			}
-		}
+		/* Wait for signals to arrive. */
+		knot_millis_sleep(startup_data.loop_wait);
 	}
 
-	finalize_ctxs(concurrent_ctxs, CTL_MAX_CONCURRENT);
-
 	if (conf()->cache.srv_dbus_event & DBUS_EVENT_RUNNING) {
 		dbus_emit_running(false);
 	}
 
-	/* Unbind the control socket. */
-	knot_ctl_unbind(ctl);
-	knot_ctl_free(ctl);
+done:
+	ctl_socket_thr_end(&sctx);
+	deinit_ctls(ctls, sock_count);
+	return ret;
 }
 
 static void print_help(void)
@@ -695,7 +550,7 @@ int main(int argc, char **argv)
 	}
 
 	/* Setup base signal handling. */
-	setup_signals();
+	signals_setup();
 
 	/* Initialize cryptographic backend. */
 	dnssec_crypto_init();
@@ -769,8 +624,18 @@ int main(int argc, char **argv)
 	drop_capabilities();
 
 	/* Activate global query modules. */
-	conf_activate_modules(conf(), &server, NULL, conf()->query_modules,
-	                      &conf()->query_plan);
+	ret = conf_activate_modules(conf(), &server, NULL, conf()->query_modules,
+	                            &conf()->query_plan);
+	if (ret != KNOT_EOK) {
+		log_fatal("failed to activate query modules (%s)", knot_strerror(ret));
+		server_wait(&server);
+		server_deinit(&server);
+		conf_free(conf());
+		dbus_close();
+		log_close();
+		dnssec_crypto_cleanup();
+		return EXIT_FAILURE;
+	}
 
 	/* Check and create PID file. */
 	unsigned long pid = pid_check_and_create();
@@ -798,7 +663,7 @@ int main(int argc, char **argv)
 
 	/* Populate zone database. */
 	log_info("loading %zu zones", conf_id_count(conf(), C_ZONE));
-	server_update_zones(conf(), &server, RELOAD_ZONES);
+	server_update_zones(conf(), &server, RELOAD_FULL);
 
 	/* Check number of loaded zones. */
 	if (knot_zonedb_size(server.zone_db) == 0) {
@@ -807,27 +672,27 @@ int main(int argc, char **argv)
 
 	stats_reconfigure(conf(), &server);
 
-	/* Start it up. */
-	log_info("starting server");
 	conf_val_t async_val = conf_get(conf(), C_SRV, C_ASYNC_START);
-	ret = server_start(&server, conf_bool(&async_val));
-	if (ret != KNOT_EOK) {
+	bool async = conf_bool(&async_val);
+
+	/* Start it up. */
+	/* In async-start mode, start answering early, otherwise in the event loop. */
+	if ((ret = server_start(&server, async)) != KNOT_EOK ||
+	    (ret = event_loop(&server, socket, daemonize, pid, async)) != KNOT_EOK) {
 		log_fatal("failed to start server (%s)", knot_strerror(ret));
+		server_stop(&server);
 		server_wait(&server);
 		stats_deinit();
-		server_deinit(&server);
-		rcu_unregister_thread();
 		pid_cleanup();
+		server_deinit(&server);
 		conf_free(conf());
+		rcu_unregister_thread();
 		dbus_close();
 		log_close();
 		dnssec_crypto_cleanup();
 		return EXIT_FAILURE;
 	}
 
-	/* Start the event loop. */
-	event_loop(&server, socket, daemonize, pid);
-
 	/* Teardown server. */
 	server_stop(&server);
 	server_wait(&server);
