@@ -1,17 +1,6 @@
-/*  Copyright (C) 2024 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+/*  Copyright (C) CZ.NIC, z.s.p.o. and contributors
+ *  SPDX-License-Identifier: GPL-2.0-or-later
+ *  For more information, see <https://www.knot-dns.cz/>
  */
 
 #include <stdio.h>
@@ -19,6 +8,7 @@
 
 #include "utils/kzonecheck/zone_check.h"
 
+#include "knot/common/log.h"
 #include "knot/zone/contents.h"
 #include "knot/zone/digest.h"
 #include "knot/zone/zonefile.h"
@@ -71,8 +61,9 @@ int zone_check(const char *zone_file, const knot_dname_t *zone_name, bool zonemd
 		.handler = { .cb = err_callback },
 	};
 
-	zloader_t zl;
-	int ret = zonefile_open(&zl, zone_file, zone_name, dflt_ttl, optional, time);
+	zloader_t loader;
+	int ret = zonefile_open(&loader, zone_file, zone_name, dflt_ttl, optional,
+	                        (sem_handler_t *)&stats, time, NULL);
 	switch (ret) {
 	case KNOT_EOK:
 		break;
@@ -80,15 +71,23 @@ int zone_check(const char *zone_file, const knot_dname_t *zone_name, bool zonemd
 	case KNOT_EFILE:
 		ERR2("failed to load the zone file");
 		return ret;
+	case KNOT_ESOAINVAL:
+		ERR2("failed to detect zone origin (missing SOA)");
+		return ret;
 	default:
 		ERR2("failed to run semantic checks (%s)", knot_strerror(ret));
 		return ret;
 	}
-	zl.err_handler = (sem_handler_t *)&stats;
-	zl.creator->master = true;
 
-	zone_contents_t *contents = zonefile_load(&zl, threads);
-	zonefile_close(&zl);
+	if (zone_name == NULL) {
+		knot_dname_txt_storage_t origin;
+		if (knot_dname_to_str(origin, loader.scanner.zone_origin, sizeof(origin)) != NULL) {
+			log_debug("detected zone origin %s", origin);
+		}
+	}
+
+	zone_contents_t *contents = zonefile_load(&loader, threads);
+	zonefile_close(&loader);
 	if (contents == NULL && !stats.handler.error) {
 		ERR2("failed to run semantic checks");
 		return KNOT_ERROR;
@@ -120,7 +119,7 @@ int zone_check(const char *zone_file, const knot_dname_t *zone_name, bool zonemd
 			fprintf(stderr, "\n");
 		}
 		printf(";; Zone dump (Knot DNS %s)\n", PACKAGE_VERSION);
-		zone_dump_text(contents, stdout, false, NULL);
+		zone_dump_text(contents, NULL, stdout, false, NULL);
 	}
 
 	zone_contents_deep_free(contents);
